import re
import os
import sys
import argparse
import glob
import uuid
import subprocess

### Herui - 2024-09

usage = "AccuSNV - SNV calling tool for bacterial isolates using deep learning."
script_dir = os.path.split(os.path.abspath(__file__))[0]


def build_dir(indir):
    if not os.path.exists(indir):
        os.makedirs(indir)


def del_same(arr1, arr2):
    if len(arr1) > 1:
        for a in arr1:
            if a in arr2:
                arr1.remove(a)
    if len(arr2) > 1:
        for a in arr2:
            if a in arr1:
                arr2.remove(a)


def findfastqfile(dr, smple, filename):
    ##### Add by Herui - 20240919 - Modified function based on the codes from Evan
    # Given the input path and filename, will return the fastq file (include SE, PE, different suffixs) Will gzip the file automatically.
    file_suffixs = ['.fastq.gz', '.fq.gz', '.fastq', '.fq',
                    '_001.fastq.gz', '_001.fq.gz', '_001.fastq', '_001.fq']
    # Check whether the file is the soft link
    target_f = []
    for f in glob.glob(f"{dr}/*"):
        # print(f,dr)
        if not os.path.islink(f):
            target_f.append(f)

    # print('target...',target_f)
    # exit()
    # Search for filename as a prefix
    files_F = [f for f in target_f if re.search(f"{filename}_?.*?R?1({'|'.join(file_suffixs)})", f)]
    files_R = [f for f in target_f if re.search(f"{filename}_?.*?R?2({'|'.join(file_suffixs)})", f)]
    # print(files_F)
    # Search for filename as a directory
    file_F = []
    file_R = []
    if os.path.isdir(f"{dr}/{filename}"):
        target_f = []
        for f in glob.glob(f"{dr}/{filename}/*"):
            if not os.path.islink(f):
                target_f.append(f)
        files_F = files_F + [f"{filename}/{f}" for f in target_f
                             if re.search(f"{filename}/.*_?.*?R?1({'|'.join(file_suffixs)})", f)]
        files_R = files_R + [f"{filename}/{f}" for f in target_f
                             if re.search(f"{filename}/.*_?.*?R?2({'|'.join(file_suffixs)})", f)]

    # print(files_F,files_R)
    del_same(files_F, files_R)
    if len(files_F) == 0 and len(files_R) == 0:
        # Can be single-end reads and no "1" or "2" ID in the filename
        print(f'No file found in {dr} for sample {smple} with prefix {filename}! Go single-end checking!')
        files_F = [f for f in target_f if re.search(f"{filename}.*_?.*({'|'.join(file_suffixs)})", f)]

        if os.path.isdir(f"{dr}/{filename}"):
            files_F = files_F + [f"{filename}/{f}" for f in target_f if
                                 re.search(f"{filename}/.*_?.*({'|'.join(file_suffixs)})", f)]
        if len(files_F) == 0:
            raise ValueError(f'No file found in {dr} for sample {smple} with prefix {filename}')
        else:
            file_F = files_F[0]
            if not file_F.endswith('.gz'):
                subprocess.run("gzip " + file_F, shell=True)
                file_F += '.gz'
        # print(files_F)

    elif len(files_F) > 1 or len(files_R) > 1:
        # print(",".join(files_F))
        # print(",".join(files_R))
        raise ValueError(f'More than 1 matching files found in {dr} for sample {smple} with prefix {filename}:\n \
                         {",".join(files_F)}\n \
                         {",".join(files_R)}')

    elif len(files_F) == 1 and len(files_R) == 1:
        file_F = files_F[0]
        file_R = files_R[0]

        ## Zip fastq files if they aren't already zipped
        if not file_F.endswith('.gz'):
            subprocess.run("gzip " + file_F, shell=True)
            file_F += '.gz'
        if not file_R.endswith('.gz'):
            subprocess.run("gzip " + file_R, shell=True)
            file_R += '.gz'
    elif len(files_F) == 1 or len(files_R) == 1:
        if len(files_F) == 1:
            file_F = files_F[0]
            if not file_F.endswith('.gz'):
                subprocess.run("gzip " + file_F, shell=True)
                file_F += '.gz'
        if len(files_R) == 1:
            file_R = files_R[0]
            if not file_R.endswith('.gz'):
                subprocess.run("gzip " + file_R, shell=True)
                file_R += '.gz'
    if file_R == []:
        file_R = ''
    if file_F == []:
        file_F = ''
    return [file_F, file_R]


def pre_check_type(infile):
    f = open(infile, 'r')
    line = f.readline()
    d = {}
    while True:
        line = f.readline().strip()
        if not line: break
        ele = re.split(',', line)
        d[ele[-1]] = ''
    if len(d) == 1:
        print('There are only SE or PE reads in the input. Use single mode!')
        return False
    else:
        print('There are both SE and PE reads in the input. Use mix mode!')
        return True


def process_input_sfile(infile, out_dir, uid, tem_dir):
    # This function will choose Snakefile according to the input type and set the soft link for input file
    # Generate new sample file for input file
    f = open(infile, 'r')
    file_prefix = os.path.splitext(os.path.basename(infile))[0]
    pre_check = pre_check_type(infile)
    # print(file_prefix)
    # exit()
    o = open(tem_dir + '/' + file_prefix + '_rebuild_' + uid + '_tem.csv', 'w+')
    line = f.readline()
    o.write(line)
    d = {}  # used to check whether the sequencing type contains 1. only SE? 2. only PE? 3. Both SE and PE.
    while True:
        line = f.readline().strip()
        if not line: break
        ele = re.split(',', line)
        files = findfastqfile(ele[0], ele[1], ele[2])

        abspath = os.path.abspath(ele[0])
        abspath_link = os.path.abspath(out_dir + '/link/')
        if pre_check:
            newpath1 = out_dir + '/link/' + ele[1] + '_1.fastq.gz'
            newpath2 = out_dir + '/link/' + ele[1] + '_2.fastq.gz'
            newpath1s = out_dir + '/link/' + ele[1] + '_1.fastq.gz'
        else:
            newpath1 = out_dir + '/link/' + ele[1] + '_1.fastq.gz'
            newpath2 = out_dir + '/link/' + ele[1] + '_2.fastq.gz'
        if ele[-1] == 'SE':
            if pre_check:
                newpath = newpath1s
            else:
                newpath = newpath1
            if not os.path.exists(newpath):
                if not files[0] == '':
                    print('ln -s ' + os.path.abspath(files[0]) + ' ' + newpath)
                    subprocess.run('ln -s ' + os.path.abspath(files[0]) + ' ' + newpath, shell=True)
                else:
                    print('ln -s ' + os.path.abspath(files[1]) + ' ' + newpath)
                    subprocess.run('ln -s ' + os.path.abspath(files[1]) + ' ' + newpath, shell=True)
            else:
                print(newpath1 + ' exists! Skip data links.')
        else:
            if not os.path.exists(newpath1):
                print('ln -s ' + os.path.abspath(files[0]) + ' ' + newpath1)
                subprocess.run('ln -s ' + os.path.abspath(files[0]) + ' ' + newpath1, shell=True)
            else:
                print(newpath1 + ' exists! Skip data links.')
            if not os.path.exists(newpath2):
                print('ln -s ' + os.path.abspath(files[1]) + ' ' + newpath2)
                subprocess.run('ln -s ' + os.path.abspath(files[1]) + ' ' + newpath2, shell=True)
            else:
                print(newpath2 + ' exists! Skip data links.')
        ele[0] = abspath_link + '/'
        o.write(','.join(ele) + '\n')
        d[ele[-1]] = ''
    o.close()
    print('New sample info file:', tem_dir + '/' + file_prefix + '_rebuild_' + uid + '_tem.csv',
          ' is generated for you. Please use it for the Snakemake pipeline!')
    return tem_dir + '/' + file_prefix + '_rebuild_' + uid + '_tem.csv'
    # choose Snakefile according to the input type
    '''
    if len(d)==1:
        if 'SE' in d:
            os.system('cp '+script_dir+'/Snakefile-3cases/Snakefile_se Snakefile')
        else:
            os.system('cp ' + script_dir + '/Snakefile-3cases/Snakefile_pe Snakefile')
    else:
        os.system('cp ' + script_dir + '/Snakefile-3cases/Snakefile_se_pe Snakefile')
    '''


def copy_config_files(sfile, uid, odir, all_p, idle_p, tf_slurm, tem_dir):
    cond = odir + '/conf'
    build_dir(cond)
    pwd=os.getcwd()
    nexp = tem_dir + '/experiment_info_' + uid + '_tem.yaml'
    os.system('cp experiment_info.yaml ' + nexp)
    o = open(cond + '/config.yaml', 'w+')
    if tf_slurm == 1:
        os.system('cp config_files/config_no_slurm.yaml ./config.yaml')
    f = open('config.yaml', 'r')
    while True:
        line = f.readline()
        if not line: break
        if re.search('configfile', line):
            line = re.sub('\./experiment_info.yaml', './' + nexp, line)
        if re.search('partition=\"', line):
            ele = re.split('\"', line)
            if len(idle_p) < 3:
                line = ele[0] + "\"" + ','.join(all_p) + "\"\n"
            else:
                line = ele[0] + "\"" + ','.join(idle_p) + "\"\n"
        o.write(line)
    o.close()
    o2 = open(pwd + '/scripts/dry_run.sh', 'w+')
    o2.write('snakemake -np  --profile ' + cond + '\n')
    o3 = open(pwd + '/scripts/run_snakemake.slurm', 'w+')
    o3.write('#!/bin/bash\n')
    o3.write('#SBATCH --job-name widevariant.main\n')
    if len(idle_p) < 3:
        o3.write('#SBATCH -n 1\n#SBATCH -p ' + ','.join(all_p) + '\n')
    else:
        o3.write('#SBATCH -n 1\n#SBATCH -p ' + ','.join(idle_p) + '\n')
    o3.write('#SBATCH --time=10:00:00\n')
    o3.write('#SBATCH --mem=50GB\n')
    o3.write('#SBATCH -o mainout.txt\n')
    o3.write('#SBATCH -e mainerr.txt\n')
    o3.write('#SBATCH --mail-user=YOUR_EMAIL_HERE\n')
    o3.write('#SBATCH --mail-type=ALL\n')
    o3.write('# Activate conda environment (may need to change name of env)\n')
    o3.write('#source activate snakemake\n')
    o3.write('snakemake  --profile ' + cond + '\n')
    o3.write('# Print "Done!!!" at end of main log file\n')
    o3.write('echo Done!!!\n')


def reset_exp_file(infile, outdir, uid, sfile, ref_dir, all_p, idle_p, tf_slurm, tem_dir, min_cov_filt, min_cov_samp,
                   exclude_samp, greport):
    f = open(infile, 'r')
    tfile = tem_dir + '/exp_' + uid + '_tem.yaml'
    o = open(tfile, 'w+')
    while True:
        line = f.readline().strip()
        if not line: break
        if re.search('outdir', line):
            o.write('outdir: ' + outdir + '\n')
        elif re.search('sample_table', line):
            o.write('sample_table: ' + sfile + '\n')
        elif re.search('ref_genome_directory', line) and not ref_dir == '':
            ref_dir = os.path.abspath(ref_dir)
            o.write('ref_genome_directory: ' + ref_dir + '\n')
        elif re.search('min_cov_samp', line):
            o.write('min_cov_samp: \"' + str(min_cov_samp) + '\"\n')
        elif re.search('min_cov_filt', line):
            o.write('min_cov_filt: \"' + str(min_cov_filt) + '\"\n')
        elif re.search('exclude_samp', line):
            o.write('exclude_samp: \"' + str(exclude_samp) + '\"\n')
        elif re.search('greport', line):
            o.write('greport: \"' + str(greport) + '\"\n')
        else:
            o.write(line + '\n')
    o.close()
    os.system(' mv ' + tfile + ' ' + infile)
    copy_config_files(sfile, uid, outdir, all_p, idle_p, tf_slurm, tem_dir)


def is_partition_valid(partition_name):
    try:
        result = subprocess.run(
            ["sinfo", "-p", partition_name],
            capture_output=True,
            text=True,
            check=False
        )

        return (result.returncode == 0 and
                len(result.stdout.split('\n')) > 1)
    except Exception:
        return False


def parse_time_limit(time_limit):
    days, hours, minutes = 0, 0, 0
    parts = time_limit.split("-")

    if len(parts) == 2:
        days = int(parts[0])
        time_part = parts[1]
    else:
        time_part = parts[0]

    time_parts = list(map(int, time_part.split(":")))

    if len(time_parts) == 3:
        hours, minutes = time_parts[:2]
    elif len(time_parts) == 2:
        minutes = time_parts[0]

    return days * 24 * 60 + hours * 60 + minutes


def is_partition_time_limit_exceed(partition):
    try:

        time_limit = subprocess.check_output(
            f"sinfo -h -o '%l' -p {partition}", shell=True, text=True
        ).strip()

        return parse_time_limit(time_limit) >= 300
    except Exception:
        return False


def snakefile_modify(infile, cp_env):
    o = open('sk_tem', 'w+')
    f = open(infile, 'r')
    lines = f.read().split('\n')
    for line in lines:
        if not re.search('source', line):
            o.write(line + '\n')
        else:
            # ele=re.split('accusnv_sub',line)
            if not re.search('source accusnv_sub', line):
                o.write(line + '\n')
            else:
                out = re.sub('accusnv_sub', cp_env, line)
                # out=ele[0]+res
                o.write(out + '\n')
    o.close()
    os.system('mv sk_tem Snakefile')


'''
def snakefile_modify_accusnv(infile,min_cov_filt,min_cov_samp,exclude_samp,greport):
	o = open('sk_tem', 'w+')
	f=open(infile,'r')
	lines = f.read().split('\n')
	for line in lines:
		if not re.search('new_snv_script',line):
			o.write(line + '\n')
		else:
			if exclude_samp=='':
				out=re.sub('-s 45 -v 5','-s '+str(min_cov_samp)+' -v '+str(min_cov_filt)+' -g '+str(greport),line)
			else:
				out = re.sub('-s 45 -v 5', '-s ' + str(min_cov_samp) + ' -v ' + str(min_cov_filt) + ' -g ' + str(greport)+' -e '+str(exclude_samp), line)
			o.write(out + '\n')
	o.close()
	os.system('mv sk_tem Snakefile')
'''


def copy_files(script_dir, pwd):
    if not os.path.exists(pwd + '/scripts'):
        os.makedirs(pwd + '/scripts')
    #if not os.path.exists(pwd + '/modules'):
    #    os.makedirs(pwd + '/modules')
    if not os.path.exists(pwd + '/CNN_models'):
        os.makedirs(pwd + '/CNN_models')
    for fname in os.listdir(script_dir + '/scripts'):
        if not os.path.exists(pwd + '/scripts/' + fname):
            os.system('cp -rf ' + script_dir + '/scripts/' + fname + ' ' + pwd + '/scripts/' + fname)
    #for fname in os.listdir(script_dir + '/modules'):
    #    if not os.path.exists(pwd + '/modules/' + fname):
    #        os.system('cp -rf ' + script_dir + '/modules/' + fname + ' ' + pwd + '/modules/' + fname)
    #if not os.path.exists(pwd+'/CNN_models/checkpoint_best_3conv.pt'):
    #    os.system('cp '+ script_dir + '/CNN_models/checkpoint_best_3conv.pt ' + pwd+'/CNN_models')
    os.system('cp ' + script_dir + '/Snakefile ' + pwd)
    os.system('cp ' + script_dir + '/experiment_info.yaml ' + pwd)
    os.system('cp ' + script_dir + '/config.yaml ' + pwd)
    os.system('cp ' + script_dir + '/experiment_info.yaml ' + pwd)
    os.system('cp ' + script_dir + '/slurm_status_script.py ' + pwd)
    #os.system('cp ' + script_dir + '/new_snv_script.py ' + pwd)
    #os.system('cp ' + script_dir + '/CNN_pred.py ' + pwd)


def main():
    pwd = os.getcwd()

    # Get para
    parser = argparse.ArgumentParser(prog='AccuSNV', description=usage)
    parser.add_argument('-i', '--input_sample_info', dest='input_sp', type=str, required=True,
                        help="The dir of input sample info file --- Required")
    parser.add_argument('-t', '--turn_off_slurm', dest='tf_slurm', type=int,
                        help="If set to 1, the SLURM system will not be used for automatic job submission. Instead, all jobs will run locally or on a single node. (Default: 0)")
    parser.add_argument('-c', '--conda_prebuilt_env', dest='cp_env', type=str,
                        help="The absolute dir of your pre-built conda env. e.g. /path/snake_pipeline/accusnv_sub")
    parser.add_argument('-r', '--ref_dir', dest='ref_dir', type=str, help="The dir of your reference genomes")
    #### AccuSNV - CNN-filter part params
    parser.add_argument('-s', '--min_cov_for_filter_sample', dest='min_cov_samp', type=str,
                        help="Before running the CNN model, low-quality samples with more than 45%% of positions having zero aligned reads will be filtered out. (default \"-s 45\") You can adjust this threshold with this parameter; to include all samples, set \"-s 100\".")
    parser.add_argument('-v', '--min_cov_for_filter_pos', dest='min_cov', type=str,
                        help="For the filter module: on individual samples, calls must have at least this many reads on the fwd/rev strands individually. If many samples have low coverage (e.g. <5), then you can set this parameter to smaller value. (e.g. -v 2). Default is 5.")
    parser.add_argument('-e', '--excluse_samples', dest='exclude_samp', type=str,
                        help="The names of the samples you want to exclude (e.g. -e S1,S2,S3). If you specify a number, such as \"-e 1000\", any sample with more than 1,000 SNVs will be automatically excluded.")
    parser.add_argument('-g', '--generate_report', dest='generate_rep', type=str,
                        help="If not generate html report and other related files, set to 0. (default: 1)")
    ####
    parser.add_argument('-o', '--output_dir', dest='out_dir', type=str,
                        help='Output dir (default: current dir/wd_out_(uid), uid is generated randomly)')  # uid=uuid.uuid1().hex
    args = parser.parse_args()
    input_file = args.input_sp
    out_dir = args.out_dir
    ref_dir = args.ref_dir
    tf_slurm = args.tf_slurm
    cp_env = args.cp_env
    # CNN-filter params
    min_cov_samp = args.min_cov_samp
    min_cov_filt = args.min_cov
    greport = args.generate_rep
    exclude_samp = args.exclude_samp

    uid = uuid.uuid1().hex
    if not out_dir:
        out_dir = pwd + '/wd_out_' + uid
    if not ref_dir:
        ref_dir = ''
    if not cp_env:
        cp_env = ''
    if not tf_slurm:
        tf_slurm = 0
    if not min_cov_filt:
        min_cov_filt = 5
    else:
        min_cov_filt = int(min_cov_filt)

    if not min_cov_samp:
        min_cov_samp = 45
    else:
        min_cov_samp = int(min_cov_samp)

    if not exclude_samp:
        exclude_samp = 'null'

    if not greport:
        greport = 1
    else:
        greport = int(greport)

    # snakefile_modify_accusnv('Snakefile',min_cov_filt,min_cov_samp,exclude_samp,greport)


    print('####script_dir:', script_dir)
    print('####current_work_dir:', pwd)
    #exit()
    copy_files(script_dir, pwd)

    if not cp_env == '':
        cp_env = os.path.abspath(cp_env)
        os.system('cp Snakefiles_diff_options/Snakefile ./')
        snakefile_modify('Snakefile', cp_env)

    tem_dir = out_dir + '/temp'
    data_dir = out_dir + '/link'
    build_dir(tem_dir)
    build_dir(data_dir)
    all_p_raw = subprocess.check_output("sinfo -h -o '%P' | sort -u", shell=True, text=True).split()
    all_p = []
    for s in all_p_raw:
        if is_partition_valid(s) and is_partition_time_limit_exceed(s):
            all_p.append(s)

    idle_p_raw = subprocess.check_output("sinfo -h -o '%P %T' | awk '$2 == \"idle\" {print $1}'", shell=True,text=True).split()
    idle_p = []
    for s in idle_p_raw:
        if s in all_p:
            idle_p.append(s)
    sfile = process_input_sfile(input_file, out_dir, uid, tem_dir)
    reset_exp_file(pwd + '/experiment_info.yaml', out_dir, uid, sfile, ref_dir, all_p, idle_p, tf_slurm, tem_dir,min_cov_filt, min_cov_samp, exclude_samp, greport)

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan  4 22:14:06 2022

@author: evanqu
"""

import numpy as np
import sys
import gzip
import argparse
import gus_helper_functions as ghf
import pickle

#%% Version history
#2022.02.08: Evan: Direct translation from pileup_to_diversity_matrix_snakemake.m
#2022.10.18, Arolyn: Now works when reference genome has lowercase letters or ambiguous letters
#2022.10.23, Arolyn: Updated comments on 40 statistics to have python indexing (0-39) as opposed to matlab indexing (1-40)

#%%Some notes

# This function saves the following 40 statistics for each position on the
# reference genome for the sample being analyzed:
# [A T C G a t c g Aq ... gq Am .... gm  At .... gt Ps Pb Pm Pftd Prtd E I D]
# List of statistics by index:
# [0-3] A is the number of forward reads supporting A
# [4-7] a is the number of reverse reads supporting A
# [8-15] Aq is the average phred qualities of all A's
# [16-23] Am is the average mapping qualities of all A's
# [24-31] At is the average tail distance of all A's
# [32] -- STATISTIC NOT COMPUTED OR RECORDED IN THIS VERSION -- 
# Ps is the p value for strand bias (fishers test)
# [33] -- STATISTIC NOT COMPUTED OR RECORDED IN THIS VERSION -- 
# Pb is the p value for the base qualities being the same for the two
# different types of calls (1st major, 2nd major nt, either strand) (ttest)
# [34] -- STATISTIC NOT COMPUTED OR RECORDED IN THIS VERSION -- 
# Pm is the p value for the mapping qualities being the same for the two
# different types of calls (ttest)
# [35] -- STATISTIC NOT COMPUTED OR RECORDED IN THIS VERSION -- 
# Pftd is the p value for the tail distantces on the forward strand
# being the same for the two different types of calls (ttest)
# [36] -- STATISTIC NOT COMPUTED OR RECORDED IN THIS VERSION -- 
# Pftd is the p value for the tail distantces on the reverse strand
# being the same for the two  different types of calls (ttest)
# [37] E is number of calls at ends of a read -- STATISTIC NOT COMPUTED OR RECORDED IN THIS VERSION -- 
# [38] I is number of reads supporting insertions in the +/- (indelregion) bp region
# [39] D is number of reads supporting deletions in the +/- (indelregion) bp region

# ChrStarts: is an array holding the indices in the position dimension
# corresponding to the start of a new chromsome.

#%%
def pileup2diversity(input_pileup, path_to_ref):
    """Grabs relevant allele info from mpileupfile and stores as a nice array 

    Args:
        input_pileup (str): Path to input pileup file.
        path_to_ref (str): Path to reference genome file
        
    """
    #Set parameters
    Phred_offset=33 #mpileup is told that the reads are in fastq format and 
                    #corrects so its always at Phred+33 when the mpileup comes out
    nts='ATCGatcg'
    nts_dict={'A':0,'T':1,'C':2,'G':3,'a':4,'t':5,'c':6,'g':7}
    num_fields=40
    indelregion=3 #region surrounding each p where indels recorded 
    #get reference genome + position information
    chr_starts,genome_length,scaf_names = ghf.genomestats(path_to_ref)
    
    #init
    data = np.zeros((genome_length,num_fields)) #format [[A T C G  a t c g],[...]]
        
    #Read in mpileup file
    print(f"Reading input file: {input_pileup}")
    mpileup = open(input_pileup)
    
    #####
    loading_bar=0
    
    for line in mpileup:

        loading_bar+=1
        if loading_bar % 50000 == 0:
            print('.')

        lineinfo = line.strip().split('\t')
        
        #holds info for each position before storing in data
        temp = np.zeros((num_fields))
        
        chromo = lineinfo[0]
        #position (absolute)
        if len(chr_starts) == 1:
            position=int(lineinfo[1])
        else:
            if chromo not in scaf_names:
                raise ValueError("Scaffold name in pileup file not found in reference")
            position=int(chr_starts[np.where(chromo==scaf_names)]) + int(lineinfo[1])
            #chr_starts starts at 0
        
        #ref allele
        ref_str = lineinfo[2]; # reference allele from pileup (usually A T C or G but sometimes a different symbol if nucleotide is ambiguous)
        if ref_str in nts_dict.keys():
            ref=nts_dict[ref_str] # convert to 0123
            if ref >= 4:
                ref = ref - 4
        else:
            ref=-1 # for cases where reference base is ambiguous
        
        #calls info
        calls=np.fromstring(lineinfo[4], dtype=np.int8) #to ASCII
        
        #qual info
        bq=np.fromstring(lineinfo[5], dtype=np.int8) # base quality, BAQ corrected, ASCII
        mq=np.fromstring(lineinfo[6], dtype=np.int8) # mapping quality, ASCII
        td=np.fromstring(lineinfo[7], dtype=int, sep=',') # distance from tail, comma sep
        
        #find starts of reads ('^' in mpileup)
        startsk=np.where(calls==94)[0]
        for k in startsk:
            calls[k:k+2]=-1
            #remove mapping character, 
            #absolutely required because the next chr could be $
        
        #find ends of reads ('$' in mpileup)
        endsk=np.where(calls==36)[0]
        calls[endsk]=-1
        
        #find indels + calls from reads supporting indels ('+-')
        indelk = np.where((calls==43) | (calls==45))[0]
        for k in indelk:
            if (calls[k+2] >=48) and (calls[k+2] < 58): #2 digit indel (size > 9 and < 100)
                indelsize=int(chr(calls[k+1]) + chr(calls[k+2])) 
                indeld=2
            else: #1 digit indel (size <= 9)
                indelsize=int(chr(calls[k+1]))
                indeld=1
            #record that indel was found in +/- indelregion nearby
            #indexing is slightly different here from matlab version
            if calls[k]==45: #deletion
                if (position-indelregion-1 >= 0) and (position+indelsize+indelregion-1 < genome_length): # if in middle of contig
                    #must store directly into data as it affects lines earlier and later
                    data[position-indelregion-1:position+indelsize+indelregion-1,39]+=1
                elif position-indelregion >= 0: # if at end of contig
                    data[position-indelregion-1:,39]+=1
                else: # if at beginning of contig
                    data[:position+indelsize+indelregion-1,39]+=1
            else: #insertion
                #insertion isn't indexed on the chromosome, no need for complex stuff
                if (position-indelregion-1 >= 0) and (position+indelregion-1 < genome_length): # if in middle of contig
                    data[position-indelregion-1:position+indelregion-1,38]+=1
                elif position-indelregion >= 0: # if at end of contig
                    data[position-indelregion-1:,38]+=1
                else: # if at beginning of contig
                    data[:position+indelregion-1,38]+=1 # indelsize->indelregion 2022.10.24 Evan and Arolyn

            #remove indel info from counting
            calls[k:(k+1+indeld+indelsize)] = -1 #don't remove base that precedes an indel
        
        #replace reference matches (.,) with their actual calls
        if ref >=0: # when reference allele is not ambiguous
            calls[np.where(calls==46)[0]]=ord(nts[ref]) #'.'
            calls[np.where(calls==44)[0]]=ord(nts[ref+4]) #','
        else: # added 2022.09.23 by Arolyn: for cases where reference allele is ambiguous, confirm there are no ,'s or .'s
            if np.any(calls==46) | np.any(calls==44):
                print( 'Line from mpileup: ' + line )
                raise ValueError('Error! Calls at this position allegedly match reference allele even though reference allele was ambiguous.')

        #index reads for finding scores
        simplecalls=calls[np.where(calls>0)[0]]
        #simplecalls is a tform of calls where each calls position
        #corresponds to its position in bq, mq, td
        
        #count how many of each nt and average scores
        for nt in range(8):
            nt_count=np.count_nonzero(simplecalls == ord(nts[nt]))
            if nt_count > 0:
                temp[nt]=nt_count
                # tem = (simplecalls == ord(nts[nt]))
                # if len(bq) < len(tem):
                #     print(lineinfo[1],' skips due to dimension problem.')
                #     continue
                try:
                    temp[nt+8]=round(np.sum(bq[simplecalls == ord(nts[nt])])/temp[nt])-Phred_offset
                    temp[nt+16]=round(np.sum(mq[simplecalls == ord(nts[nt])])/temp[nt])-33
                    temp[nt+24]=round(np.sum(td[simplecalls == ord(nts[nt])])/temp[nt])
                except:
                    print(lineinfo[1],' error at line 186-188! Skip!')

        
        #-1 is needed to turn 1-indexed positions to python 0-indexed
        data[position-1,:38]=temp[:38]
        
    #######
    mpileup.close()
    
    #calc coverage
    coverage=np.sum(data,1)
    
    return data, coverage

#%%
if __name__ == "__main__":
    
    SCRIPTS_DIR="scripts"
    sys.path.insert(0, SCRIPTS_DIR)
        
    parser = argparse.ArgumentParser()
    
    parser.add_argument('-i', dest='input', type=str, help='Path to input pileup',required=True)
    parser.add_argument('-r', dest='ref', type=str, help='Path to reference genome',required=True)
    parser.add_argument('-o', dest='output', type=str, help='Path to output diversity file', required=True)
    parser.add_argument('-c', dest='coverage', type=str, help='Path to coverage file', required=True)
    
    args = parser.parse_args()
    
    diversity_arr, coverage_arr = pileup2diversity(args.input,args.ref)
    
    with gzip.open(args.output, 'wb') as f:
        pickle.dump(diversity_arr,f)
    
    if args.coverage:
        with gzip.open(args.coverage, 'wb') as f:
            pickle.dump(coverage_arr,f)



"""TDL Nov 2011 edited from Nature Genetics project
edited August 2013 to just generate .tree files

"""

import os
import re
import sys
import numpy as np
import pandas as pd
import matplotlib.mlab as mlab
from pylab import *
import random
from matplotlib.font_manager import FontProperties
from scipy import stats
from datetime import datetime, date, time
from matplotlib import rc
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord



rc('font', **{'sans-serif':'Arial'})


def mutation_count(inputtree, lca):
	#print(inputtree,lca)
	#exit()
	emptycount=0
	ambiguous=['R','Y','M','K','S','W']
	ACGT=['A','C','T','G']

	tree=''
	for i,c in enumerate(inputtree):
		if c in ['(', ',', ')']:
			tree+=c
		elif c =='*':
			tree+=inputtree[i+1]
	#print(tree)
	#exit()
	#remove all question marks and ambiguities from the tree

	#simplify tree while counting mutations. give preference to simplifying first.
	simplified=1
	mutlist={('A','C'):0,('G','C'):0,('T','C'):0,('A','G'):0,('C','G'):0,('T','G'):0, ('A','T'):0,('G','T'):0,('C','T'):0, ('T','A'):0,('G','A'):0,('C','A'):0}
	dcount={}
	all_val_mut=[]
	for i in tree:
		if i=='N' or i==',' or i==')' or i=='(':continue
		if i not in dcount:
			dcount[i]=1
		else:
			dcount[i]+=1
		all_val_mut.append(i)
	res=sorted(dcount.items(), key = lambda kv:(kv[1], kv[0]), reverse = True)
	#print(res)
	#print(all_val_mut)
	anc=res[0][0]

	for i in res:
		if i[0] ==anc:continue
		mutlist[anc,i[0]]+=i[1]

	
	muts=0
	for pair in mutlist.keys():
		muts+=mutlist[pair]
	
	if muts==0:
		print(tree)
	
	return muts, mutlist


def load_chr(chart):
	f=open(chart,'r')
	line=f.readline()
	chromosomes=[]
	d={}
	while True:
		line=f.readline().strip()
		if not line:break
		ele=re.split(',',line)
		if ele[0] not in d:
			d[ele[0]]=''
			chromosomes.append(ele[0])
	return chromosomes


def convert_csv(chart_raw,mode,out):
	# if mode == 1 -> used in the Snakemake pipeline, mode==2 -> used in the downstream analysis module
	chart = out + '/snpChart.csv'
	df = pd.read_csv(chart_raw, sep='\t')
	c=0
	raw_pos=[]
	keep_col=['contig_idx','contig_pos']
	#print(df.columns)
	#exit()
	for col in df.columns:
		if c == 0:
			c += 1
			raw_pos = np.array(df[col])
			continue
		if mode==1:
			if c < 34:
				# if col=='contig_idx' or col=='contig_pos':
				# 	keep_col.append(col)
				c+=1
				continue
		else:
			if c < 19:
				c+=1
				continue

		if re.search('sequence', col): continue
		if re.search('transl', col): continue
		if re.search('Unnamed', col): continue
		keep_col.append(col)
	#print(keep_col)
	#exit()
	subset=df[keep_col]
	subset.rename(columns={'contig_pos': 'pos'}, inplace=True)
	subset.rename(columns={'contig_idx': 'chr'}, inplace=True)
	subset.to_csv(chart, index=False)

	return chart,raw_pos
		


def mutationtypes(tree, chart_raw,mode,out):

	chart,raw_pos=convert_csv(chart_raw,mode,out)
	#print(chart,raw_pos)
	#exit()


	tout=out+'/snp_trees'
	if not os.path.exists(tout):
		os.makedirs(tout)
		

	ATCG=['A','C','T','G']


		
	f=open(chart,'r').readlines()
	#chromosomes=load_chr(chart)

	
	pidx=0
	for i, line in enumerate(f[1:]):
	
		l=line.strip().split(',')
		if len(l) < 5:
			print(l)
			pidx+=1
			continue
		chromosome=l[0]
		#print(chromosome)
		#exit()
		pos=l[1]
		
		#use first strain as lca
		lca=l[2]


		#count mutations
		newtree = annotateSNP(tree, chart, chromosome, pos)

		a, mutlist= mutation_count(newtree, lca)
		#print(a,mutlist)
		#exit()

		if a==0:
			print('NO MUTS:')
			#print(len(line), i, chromosome, pos)
		#save trees
		else:
			f1=open(tout+'/p_'+str(raw_pos[pidx])+'_'+str(a)+'.tree','w')
			#f1=open(tout+'/'+'p_'+str(raw_pos[pidx])+'.tree','w')
			t=open('tempnexus.txt','r').readlines()
			for line in t:
				f1.write(line)
		pidx+=1
	os.system('rm temptree.txt')
	os.system('rm tempnexus.txt')



def annotateSNP(tree, dict, chromosome,pos):
	#chromosomes=['NZ_CH482380.1','NZ_CH482381.1','NZ_CH482382.1']

	f=open(dict).readlines()
	fo=open('temptree.txt','w')
	
	header=f[0].strip().split(',')
	locations=[]
	
	for n,i in enumerate(header):
		if not re.search('pos',i) and not re.search('chr',i) :
			locations.append(n)
	#print(locations)
	#exit()

	for line in f:
		l=line.strip('\n').split(',')
		#print(l,chromosome,pos)
		#exit()
		if l[0]==chromosome and l[1]==pos:
			for i in locations:
				#print(len(l),i)
				#exit()
				if len(l) > i and len(l[i])>0:
					fo.write(header[i]+'\t'+l[i]+'\n')
				else:
					fo.write(header[i]+'\t?\n')
				if i > len(l):
					print(line, l, i)
			break
	#exit()
	fo.close()
	newtree = nameswap(tree,'temptree.txt')
	return newtree



def nameswap(tree, dictionary):

	fraw=open(dictionary).readlines()
	f=[]
	for r in fraw:
		if not re.search('muts\t',r) and not re.search('type\t',r):
			f.append(r)
	#print(f,len(f))
	#exit()
	numStrains=len(f)
	dict={}
	annotation={}
	newname={}
	
	#print header for tempnexus
	fo=open('tempnexus.txt','w')
	fo.write('#NEXUS\nbegin taxa;\n\tdimensions ntax='+str(numStrains)+';\n\ttaxlabels\n')			
	colors={'A':'[&!color=#-16776961]', 'C':'[&!color=#-16725916]','G':'[&!color=#-3670016]','T':'[&!color=#-3618816]','?':'[&!color=#-16777216]','N':'[&!color=#-16777216]'}
	ambiguous=['R','Y','M','K','S','W']
	for each in ambiguous:
		colors[each]='[&!color=#-16777216]'
				
	#get annotations
	f=open(dictionary, 'r').readlines()
	for line in f:
		if not line.startswith('#'):
			l=line.split()
			annotation[l[0]]=l[1]
	
	
	#combine names and annotations	
	for i in annotation.keys():
		newname[i]=i+ '--*'+ annotation[i]
		#print(newname)
		#exit()
		#if i in dict.keys():
			#newname[i]=dict[i]+ '--*'+ annotation[i] #for dating newname[i]=dict[i]+ '--*'+ annotation[i]
		#else:
			#newname[i]= i + '--*'+ annotation[i] #for reference, etc.
	#print(newname)
	#exit()
	#print(dictionary,annotation.keys())
	#exit()
	#make new tree
	f=open(tree,'r').readlines()
	#print(tree)
	
	newtree=''
	
	for line in f:
		line=line.strip()
		#print(line)
		line=re.sub(',reference_genome:0.00000,inferred_ancestor:0.00000','',line)
		i=0
		#print(line)
		# Regular expression to extract key:value pairs
		pattern = r"([\w\-]+:\d+\.\d+)"

		# Find all matches in the string
		matches = re.findall(pattern, line)
		#print(matches)
		#exit()

		# Create an empty list to store the replaced patterns
		replaced_patterns = []
		#print(matches)
		nline=line
		# Replace keys using the dictionary and store the results
		for match in matches:
			key, value = match.split(":")
			#print(key,value)
			if key=='reference_genome' or key=='inferred_ancestor' or newname[key][-1] not in colors:
				print(key)
				continue
			if key in newname:
				# Replace the key with the corresponding value from the dictionary
				new_key = newname[key]
				replaced_patterns.append(f"{new_key}:{value}")
				#new = line.replace(key, new_key)
				#newtree += new
				fo.write('\t\'' + new_key + '\'' + colors[new_key[-1]] + '\n')

				nline=re.sub(key+':',newname[key]+':',nline)

		#print(replaced_patterns)
		#print(replaced_patterns)
		#exit()

		newtree+=nline
		#print(newtree)
		#exit()
		'''
		for key, value in newname.items():
			if key in line:
				new = line.replace(key, value)
				fo.write('\t\'' + new + '\'' + colors[value[-1]] + '\n')
		
		while i < len(line):
			if line[i] not in ['T']: #T for TL...
				newtree+=line[i]
				i+=1
			else:
				oldname=line[i:i+5]
				i=i+5
				if oldname in newname.keys():
					new=newname[oldname]
				else:
					new=oldname
				print(new)
				exit()
				if new[-1]=='N':
					new=new[:-1]+'?'

				print(new)
				newtree+=new
				fo.write('\t\''+new+'\''+colors[new[-1]]+'\n') #write down new colors to nexus file
		'''
	#exit()
	#write rest of nexus file
	fo.write(';\nend\n\nbegin trees;\n\ttree tree_1=[&R] ')
	for line in newtree:
		#line=re.sub(',reference_genome:0.00000,inferred_ancestor:0.00000','',line)
		#print(line)
		fo.write(line)
	#exit()
	fo.write('end;\n\nbegin figtree;\n')
	fo.write('\tset tipLabels.fontSize=10;\n')
	fo.write('\tset rectilinearLayout.curvature=7000;\n')
	fo.write('\tset layout.expansion=700;\n')
	fo.write('end;')
	fo.close()
	#exit()
	return newtree	







# if __name__ == "__main__":
# 		print('WARNING: Uses first strain as LCA')
# 		#out='tree_test'
# 		dir_output='../kcp_science_local_new'
#
# 		mutationtypes(dir_output+"/snv_tree_genome_latest.nwk.tree",dir_output+'/snv_table_merge_all_mut_annotations.tsv',dir_output)
# 		print('Done.')


		
		




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module containing helper scripts for GUS
"""
from Bio import SeqIO
from itertools import compress
import os
import re
import numpy as np
import pickle
import csv
import glob
import subprocess
import gzip

def read_samples_CSV(spls):
    hdr_check = ['Path','Sample','FileName','Reference','Group','Outgroup','Type']
    switch = "on"
    file = open(spls, 'r')
    #initialize lists
    list_path = []; list_splID = []; list_fileN = []; list_refG = []
    list_group=[]; list_outgroup=[]; list_seqtype=[]
    for line in file:
        line = line.strip('\n').split(',')
        # Test Header. Note: Even when header wrong code continues (w/ warning), but first line not read.
        if switch == "on":
            if (line == hdr_check):
                print("Passed CSV header check")
            else:
                Warning("CSV did NOT pass header check! Code continues, but first line ignored")
            switch = "off"
            continue
        # build lists
        list_path.append(line[0].strip())
        list_splID.append(line[1].strip())
        list_fileN.append(line[2].strip())
        list_refG.append(line[3].strip())
        list_group.append(line[4].strip())
        list_outgroup.append(line[5].strip())
        list_seqtype.append(line[6].strip())

    return [list_path,list_splID,list_fileN,list_refG,list_group,list_outgroup,list_seqtype]


def split_samplesCSV(PATH_ls,SAMPLE_ls,FILENAME_ls,REF_Genome_ls,GROUP_ls,OUTGROUP_ls,TYPE_ls,outdir):
    '''Take info from samples.csv, concat by sample name & save each line as sample_info.csv in data/{sampleID}'''
    
    #Loop through unique samples
    for sample in set(SAMPLE_ls):
        # Concat info for this sample
        sample_info_bool=[s==sample for s in SAMPLE_ls]
        sample_paths=list(compress(PATH_ls,sample_info_bool))
        sample_filenames=list(compress(FILENAME_ls,sample_info_bool))
        sample_references=list(compress(REF_Genome_ls,sample_info_bool))
        sample_groups=list(compress(GROUP_ls,sample_info_bool))
        sample_outgroups=list(compress(OUTGROUP_ls,sample_info_bool))
        sample_type = list(compress(TYPE_ls, sample_info_bool))
        
        sample_info_csv=list(zip(sample_paths,[sample]*sum(sample_info_bool),sample_filenames,sample_references,sample_groups,sample_outgroups,sample_type))
        
        # make data directory for this sample if it doesn't already exist
        if not(os.path.isdir(outdir+'/data/' + sample)):
            os.makedirs(outdir+'/data/' + sample, exist_ok=True)
        # check to see if this mini csv with sample info already exists
        if os.path.isfile(outdir+'/data/' + sample + '/sample_info.csv'):
            # if so, read file
            old_file = open(outdir+'/data/' + sample + '/sample_info.csv','r')
            old_info_read = csv.reader(old_file)
            old_info = list(map(tuple, old_info_read))
            old_file.close()
            
            # check to see if the existing file is consistent with samples.csv
            if not(old_info == sample_info_csv):
                # if not, remove the old file and save sample info in a new file
                # print('Information file must be updated.')
                os.remove(outdir+'/data/' + sample + '/sample_info.csv')
                with open(outdir+'/data/' + sample + '/sample_info.csv', "w") as f:
                    writer = csv.writer(f)
                    for row in sample_info_csv:
                        writer.writerow(row)
        else: # if mini csv with sample info does not already exist
            # save sample info in mini csv
            with open(outdir+'/data/' + sample + '/sample_info.csv', "w") as f:
                writer = csv.writer(f)
                for row in sample_info_csv:
                    writer.writerow(row)



def findfastqfile(dr, smple, filename):
    ##### Add by Herui - 20240921 - Modified function based on the codes from Evan
    # Given the input path and filename, will return the fastq file (include SE, PE, different suffixs) Will gzip the file automatically.
    file_suffixs = ['.fastq.gz', '.fq.gz', '.fastq', '.fq',
                    '_001.fastq.gz', '_001.fq.gz', '_001.fastq', '_001.fq']
    # Check whether the file is the soft link
    target_f=[]
    for f in glob.glob(f"{dr}/*"):
        if not os.path.islink(f):
            target_f.append(f)
        else:
            link_abs_path = os.path.abspath(f)
            link_name = os.path.basename(f)
            target_f.append(link_abs_path+'/'+link_name)
            
    #print('target...',target_f)
    #exit()
    # Search for filename as a prefix
    files_F = [f for f in target_f if re.search(f"{filename}_?.*?R1({'|'.join(file_suffixs)})", f)]
    files_R = [f for f in target_f if re.search(f"{filename}_?.*?R2({'|'.join(file_suffixs)})", f)]
    #print(dr,smple,filename,glob.glob(f"{dr}/*"))
    # Search for filename as a directory
    file_F = []
    file_R = []
    if os.path.isdir(f"{dr}/{filename}"):
        target_f = []
        for f in glob.glob(f"{dr}/{filename}/*"):
            if not os.path.islink(f):
                target_f.append(f)
            else:
                link_abs_path = os.path.abspath(f)
                link_name = os.path.basename(f)
                target_f.append(link_abs_path+'/'+link_name)
                
        files_F = files_F + [f"{filename}/{f}" for f in target_f
                             if re.search(f"{filename}/.*_?.*?R1({'|'.join(file_suffixs)})", f)]
        files_R = files_R + [f"{filename}/{f}" for f in target_f
                             if re.search(f"{filename}/.*_?.*?R2({'|'.join(file_suffixs)})", f)]
    #print(files_F,files_R)
    if len(files_F) == 0 and len(files_R) == 0:
        # Can be single-end reads and no "1" or "2" ID in the filename
        #print(f'No file found in {dr} for sample {smple} with prefix {filename}! Go single-end checking!')
        files_F = [f for f in target_f if re.search(f"{filename}.*_?.*({'|'.join(file_suffixs)})", f)]

        if os.path.isdir(f"{dr}/{filename}"):
            files_F = files_F + [f"{filename}/{f}" for f in target_f if
                                 re.search(f"{filename}/.*_?.*({'|'.join(file_suffixs)})", f)]
        if len(files_F) == 0:
            raise ValueError(f'No file found in {dr} for sample {smple} with prefix {filename}')
        else:
            file_F = files_F[0]
            if not file_F.endswith('.gz'):
                subprocess.run("gzip " + file_F, shell=True)
                file_F += '.gz'
        # print(files_F)

    elif len(files_F) > 1 or len(files_R) > 1:
        # print(",".join(files_F))
        # print(",".join(files_R))
        raise ValueError(f'More than 1 matching files found in {dr} for sample {smple} with prefix {filename}:\n \
                         {",".join(files_F)}\n \
                         {",".join(files_R)}')

    elif len(files_F) == 1 and len(files_R) == 1:
        file_F = files_F[0]
        file_R = files_R[0]

        ## Zip fastq files if they aren't already zipped
        if not file_F.endswith('.gz'):
            subprocess.run("gzip " + file_F, shell=True)
            file_F += '.gz'
        if not file_R.endswith('.gz'):
            subprocess.run("gzip " + file_R, shell=True)
            file_R += '.gz'
    elif len(files_F) == 1 or len(files_R) == 1:
        if len(files_F) == 1:
            file_F = files_F[0]
            if not file_F.endswith('.gz'):
                subprocess.run("gzip " + file_F, shell=True)
                file_F += '.gz'
        if len(files_R) == 1:
            file_T = files_R[0]
            if not file_R.endswith('.gz'):
                subprocess.run("gzip " + file_R, shell=True)
                file_R += '.gz'
    if file_R==[]:
        file_R=''
    return [file_F, file_R]

def findfastqfile_old(dr,ID,filename):
    fwd=[]
    rev=[]
    #search for filename as directory first
    potentialhits_forward=glob.glob(dr + '/' + filename +'/*1.fastq.gz')
    potentialhits_reverse=glob.glob(dr + '/' + filename +'/*2.fastq.gz')
    if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
        fwd=potentialhits_forward[0]
        rev=potentialhits_reverse[0]
    #then search for filename as file.gz
    elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
        potentialhits_forward=glob.glob(dr + '/' + filename +'*1.fastq.gz')
        potentialhits_reverse=glob.glob(dr + '/' + filename +'*2.fastq.gz')
        if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
            fwd=potentialhits_forward[0]
            rev=potentialhits_reverse[0]
        #then search as unzipped file
        elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
            potentialhits_forward=glob.glob(dr + '/' + filename +'*1.fastq')
            potentialhits_reverse=glob.glob(dr + '/' + filename +'*2.fastq')
            if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                subprocess.run("gzip " + potentialhits_forward[0], shell=True)  
                subprocess.run("gzip " + potentialhits_reverse[0], shell=True)
                fwd=potentialhits_forward[0]+'.gz'
                rev=potentialhits_reverse[0]+'.gz'
            else:
                foldername=glob.glob(dr + '/' + filename + '*')
                if foldername and os.path.isdir(foldername[0]):
                    foldername=foldername[0]
                    potentialhits_forward=glob.glob(foldername + '/*' + filename + '*1*.fastq.gz')
                    potentialhits_reverse=glob.glob(foldername + '/*' + filename + '*2*.fastq.gz')
                    if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                        fwd=potentialhits_forward[0]
                        rev=potentialhits_reverse[0]
                    elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
                        print(foldername + '/*' + filename + '*2*.fastq.gz')
                        potentialhits_forward=glob.glob(foldername +  '/*' + filename + '*1*.fastq')
                        potentialhits_reverse=glob.glob(foldername + '/*' + filename + '*2*.fastq')
                        if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                            subprocess.run("gzip " + potentialhits_forward[0], shell=True)  
                            subprocess.run("gzip " + potentialhits_reverse[0], shell=True)
                            fwd=potentialhits_forward[0]+'.gz'
                            rev=potentialhits_reverse[0]+'.gz'
    if not(fwd) or not(rev):
        raise ValueError('Either no file or more than 1 file found in ' + dr + 'for ' + ID)
    ##zip fastq files if they aren't already zipped
    subprocess.run("gzip " + fwd, shell=True)   
    subprocess.run("gzip " + rev, shell=True)   
    return [fwd, rev]


def read_sample_info_CSV(path_to_sample_info_csv):
    with open(path_to_sample_info_csv, 'r') as f:
        this_sample_info = f.readline()  # only one line to read
    this_sample_info = this_sample_info.strip('#').split(',')
    path = this_sample_info[0]  # remember python indexing starts at 0
    paths = path.split(' ')
    sample = this_sample_info[1]
    filename = this_sample_info[2]
    reference = this_sample_info[3]
    stype = this_sample_info[-1]
    suff=findfastqfile(path,sample,sample)
    #print(suff)
    suff[0]=os.path.basename(suff[0])
    suff[1]=os.path.basename(suff[1])
    #print(suff)
    #exit()

    return paths, sample, reference, filename, stype,suff
#Jonathan new code
# def findfastqfile(dr,ID,filename):
#     fwd=[]
#     rev=[]
#     potentialhits_forward=glob.glob(dr + '/' + filename +'/*1.fastq.gz')
#     potentialhits_reverse=glob.glob(dr + '/' + filename +'/*2.fastq.gz')
#     if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
#         fwd=potentialhits_forward[0]
#         rev=potentialhits_reverse[0]
#     elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
#         potentialhits_forward=glob.glob(dr + '/' + filename +'*1.fastq.gz')
#         potentialhits_reverse=glob.glob(dr + '/' + filename +'*2.fastq.gz')
#         if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
#             fwd=potentialhits_forward[0]
#             rev=potentialhits_reverse[0]
#         elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
#             potentialhits_forward=glob.glob(dr + '/' + filename +'*1.fastq')
#             potentialhits_reverse=glob.glob(dr + '/' + filename +'*2.fastq')
#             if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
#                 subprocess.run("gzip " + potentialhits_forward[0], shell=True)
#     # wildcards: sampleID=20221
#     # resources: mem_mb=1000, disk_mb=1000, tmpdir=/tmp
#                 subprocess.run("gzip " + potentialhits_reverse[0], shell=True)
#                 fwd=potentialhits_forward[0]+'.gz'
#                 rev=potentialhits_reverse[0]+'.gz'
#             else:
#                 foldername=glob.glob(dr + '/' + filename + '*')
#                 if foldername and os.path.isdir(foldername[0]):
#                     foldername=foldername[0]
#                     potentialhits_forward=glob.glob(foldername + '/*' + filename + '*1*.fastq.gz')
#                     potentialhits_reverse=glob.glob(foldername + '/*' + filename + '*2*.fastq.gz')
#                     if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
#                         fwd=potentialhits_forward[0]
#                         rev=potentialhits_reverse[0]
#                     elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
#                         print(foldername + '/*' + filename + '*2*.fastq.gz')
#                         potentialhits_forward=glob.glob(foldername +  '/*' + filename + '*1*.fastq')
#                         potentialhits_reverse=glob.glob(foldername + '/*' + filename + '*2*.fastq')
#                         if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
#                             subprocess.run("gzip " + potentialhits_forward[0], shell=True)
#                             subprocess.run("gzip " + potentialhits_reverse[0], shell=True)
#                             fwd=potentialhits_forward[0]+'.gz'
#                             rev=potentialhits_reverse[0]+'.gz'
#     if not(fwd) or not(rev):
#         raise ValueError('Either no file or more than 1 file found in ' + dr + 'for ' + ID)
#     ##make read only --- addition 20220415 by Jonathan (jdg)
#     subprocess.run("chmod 444 " + fwd, shell=True)
#     subprocess.run("chmod 444 " + rev, shell=True)
#     ##zip fastq files if they aren't already zipped
#     subprocess.run("gzip " + fwd, shell=True)
#     subprocess.run("gzip " + rev, shell=True)
#     return [fwd, rev]

def makelink(path,sample,filename,output_dir):
    #When sample is run on a single lane
    #File name can be either a COMPLETE directory name or a file name in batch(called path in this fx)
    [fwd_file, rev_file]=findfastqfile(path,sample, filename)
    # subprocess.run('ln -s -T ' + fwd_file + ' data/' + sample + '/R1.fq.gz', shell=Trueq)    
    # subprocess.run('ln -s -T ' + rev_file + ' data/' + sample + '/R2.fq.gz', shell=True)    
    
    subprocess.run(f"ln -s -T {fwd_file} {output_dir}/{sample}/R1.fq.gz", shell=True)    
    subprocess.run(f"ln -s -T {rev_file} {output_dir}/{sample}/R2.fq.gz", shell=True)    


def cp_append_files(paths,sample,filename,output_dir):
    #When sample is run on multiple lanes with same barcode
    fwd_list=''
    rev_list=''
    for path in paths:
        #Provider name can be either a COMPLETE directory name or a file name in batch(called path in this fx)
        [fwd_file, rev_file]=findfastqfile(path,sample, filename)
        fwd_list=fwd_list+ ' ' + fwd_file
        rev_list=rev_list+ ' ' + rev_file
        print(rev_list)
        print(fwd_list)
    subprocess.run("zcat " + fwd_list + ' | gzip > ' + output_dir + '/' +  sample + '/R1.fq.gz', shell=True)
    subprocess.run("zcat " + rev_list + ' | gzip > ' + output_dir + '/' +  sample + '/R2.fq.gz', shell=True)
    

def read_fasta(REFGENOME_DIR): 
    '''Reads in fasta file. If directory is given, reads in dir/genome.fasta
    Args:
        REFGENOME_DIR (str): Path to reference genome.

    Returns: SeqIO object for reference genome.
    '''
    fasta_file = glob.glob(REFGENOME_DIR + '/genome.fasta')
    if len(fasta_file) != 1:
        fasta_file_gz = glob.glob(REFGENOME_DIR + '/genome.fasta.gz')
        if len(fasta_file_gz) != 1:
            raise ValueError('Either no genome.fasta(.gz) or more than 1 genome.fasta(.gz) file found in ' + REFGENOME_DIR)
        else: # genome.fasta.gz
            refgenome = SeqIO.parse(gzip.open(fasta_file_gz[0], "rt"),'fasta')
    else: # genome.fasta
        refgenome = SeqIO.parse(fasta_file[0],'fasta')
    
    return refgenome

def genomestats(REFGENOMEFOLDER):
    '''Parse genome to extract relevant stats

    Args:
        REFGENOMEFOLDER (str): Directory containing reference genome file.

    Returns:
        ChrStarts (arr): DESCRIPTION.
        Genomelength (arr): DESCRIPTION.
        ScafNames (arr): DESCRIPTION.

    '''

    refgenome = read_fasta(REFGENOMEFOLDER)
    
    Genomelength = 0
    ChrStarts = []
    ScafNames = []
    for record in refgenome:
        ChrStarts.append(Genomelength) # chr1 starts at 0 in analysis.m
        Genomelength = Genomelength + len(record)
        ScafNames.append(record.id)
    # close file
    #refgenome.close() # biopy update SeqIO has no close attribute anymore.
    # turn to np.arrys!
    ChrStarts = np.asarray(ChrStarts,dtype=int)
    Genomelength = np.asarray(Genomelength,dtype=int)
    ScafNames = np.asarray(ScafNames,dtype=object)
    return ChrStarts,Genomelength,ScafNames

def p2chrpos(p, ChrStarts):
    '''Convert 1col list of pos to 2col array with chromosome and pos on chromosome

    Args:
        p (TYPE): DESCRIPTION.
        ChrStarts (TYPE): DESCRIPTION.

    Returns:
        chrpos (TYPE): DESCRIPTION.

    '''
        
    # get chr and pos-on-chr
    chromo = np.ones(len(p),dtype=int)
    if len(ChrStarts) > 1:
        for i in ChrStarts[1:]:
            chromo = chromo + (p > i) # when (p > i) evaluates 'true' lead to plus 1 in summation. > bcs ChrStarts start with 0...genomestats()
        positions = p - ChrStarts[chromo-1] # [chr-1] -1 due to 0based index
        chrpos = np.column_stack((chromo,positions))
    else:
        chrpos = np.column_stack((chromo,p))
    return chrpos


def p2chrpos_add_base(p, ChrStarts, base_type):
    '''Convert 1col list of pos to 2col array with chromosome and pos on chromosome

    Args:
        p (TYPE): DESCRIPTION.
        ChrStarts (TYPE): DESCRIPTION.

    Returns:
        chrpos (TYPE): DESCRIPTION.

    '''

    # get chr and pos-on-chr
    chromo = np.ones(len(p), dtype=int)
    if len(ChrStarts) > 1:
        for i in ChrStarts[1:]:
            chromo = chromo + (
                        p > i)  # when (p > i) evaluates 'true' lead to plus 1 in summation. > bcs ChrStarts start with 0...genomestats()
        positions = p - ChrStarts[chromo - 1]  # [chr-1] -1 due to 0based index
        chrpos = np.column_stack((chromo, positions,base_type))
    else:
        chrpos = np.column_stack((chromo, p,base_type))
    return chrpos

# def get_clade_wildcards(cladeID):
#     is_clade = [int(i == cladeID) for i in GROUP_ls]
#     sampleID_clade = list(compress(SAMPLE_ls,is_clade))
#     reference_clade = list(compress(REF_Genome_ls,is_clade))
#     outgroup_clade = list(compress(OUTGROUP_ls,is_clade))
#     return sampleID_clade,reference_clade,outgroup_clade
    
# def get_sampleID_names(wildcards):  
#     sampleID_clade,_,_ = get_clade_wildcards(wildcards.cladeID)
#     return sampleID_clade

# def get_outgroup_bool(wildcards):  
#     _,_,outgroup_clade = get_clade_wildcards(wildcards.cladeID)
#     return outgroup_clade

# def get_positions_prep(wildcards):
#     sampleID_clade,reference_clade,outgroup_clade = get_clade_wildcards(wildcards.cladeID)
#     mat_positions_prep=expand("Case/temp/{sampleID}_ref_{reference}_outgroup{outgroup}_positions.pickle",zip,sampleID=sampleID_clade, reference=reference_clade, outgroup=outgroup_clade)
#     return mat_positions_prep

# def get_diversity(wildcards):
#     sampleID_clade,reference_clade,outgroup_clade = get_clade_wildcards(wildcards.cladeID)
#     diversity_mat = expand("Mapping/diversity/{sampleID}_ref_{reference}_outgroup{outgroup}.diversity.pickle.gz",zip,sampleID=sampleID_clade, reference=reference_clade, outgroup=outgroup_clade)
#     return diversity_mat   

# def get_quals(wildcards):
#     sampleID_clade,reference_clade,outgroup_clade = get_clade_wildcards(wildcards.cladeID)
#     quals_mat = expand("Mapping/quals/{sampleID}_ref_{reference}_outgroup{outgroup}.quals.pickle.gz",zip,sampleID=sampleID_clade, reference=reference_clade, outgroup=outgroup_clade)
#     return quals_mat 

# def get_ref_genome(wildcards):
#     sampleID_clade,reference_clade,outgroup_clade = get_clade_wildcards(wildcards.cladeID)
#     return reference_clade



import os
import re
import sys
import copy
import torch
import random
import numpy as np
import torch.nn as nn
import torch.optim as optim
from scipy import stats
from scipy.stats import norm
from statsmodels.stats.power import TTestPower
from torch.utils.data import Dataset, DataLoader
#from sklearn.metrics import accuracy_score,precision_score,recall_score,f1_score,roc_auc_score,confusion_matrix, classification_report
script_dir = os.path.dirname(os.path.abspath(__file__))
dir_py_scripts = script_dir+"/modules"
sys.path.insert(0, dir_py_scripts)
import snv_module_recoded_new as snv # SNV calling module

def setup_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic=True


# Custom Dataset Class
class CustomDataset(Dataset):
    def __init__(self, data, labels):
        self.data = torch.tensor(data, dtype=torch.float32)
        self.labels = torch.tensor(labels, dtype=torch.float32)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx], self.labels[idx]

# Make major, minor, other_1, other_2 as the channel
class CNNModel(nn.Module):
    def __init__(self, n_channels, num_classes=1):
        super(CNNModel, self).__init__()

        # 1x1 Convolution to capture channel information
        # self.conv1x1 = nn.Conv2d(in_channels=n_channels, out_channels=32, kernel_size=1)

        self.conv1 = nn.Conv2d(in_channels=n_channels, out_channels=32, kernel_size=(3,4), padding=(1, 1))
        self.conv2 = nn.Conv2d(in_channels=32, out_channels=64, kernel_size=2, padding=(1, 0))
        self.conv3 = nn.Conv2d(in_channels=64, out_channels=128, kernel_size=2, padding=(1, 0))
        # self.conv3 = nn.Conv2d(in_channels=128, out_channels=256, kernel_size=2, padding=(1, 0))

        self.global_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc1 = nn.Linear(128, 64)
        self.fc2 = nn.Linear(64, num_classes)
        self.dropout = nn.Dropout(p=0.5)

    def forward(self, x):
        # Input shape: (batch_size, sample_num, features, n_channels->ATGC)
        # print(x.shape)
        x = x.permute(0, 3, 1, 2)  # Rearrange to (batch_size, n_channels, x_dim, y_dim)
        # print(x.shape)
        # x = torch.relu(self.conv1x1(x))
        x = torch.relu(self.conv1(x))
        # print(x.shape)
        # exit()
        x = torch.relu(self.conv2(x))
        # print(x.shape)
        x = torch.relu(self.conv3(x))
        # print(x.shape)
        # exit()

        x = self.global_pool(x)
        x = x.view(x.size(0), -1)  # Flatten
        x = torch.relu(self.fc1(x))
        x = self.dropout(x)
        x = torch.sigmoid(self.fc2(x))

        # exit()
        return x

def cal_med_cov_for_given_array(x):
    nx=x.transpose(1, 0, 2)
    data=x.reshape(nx.shape[0],nx.shape[1]*nx.shape[2])
    non_zero_data = [row[row != 0] for row in data]
    median_values = np.array([np.median(row) if len(row) > 0 else 0 for row in non_zero_data])
    return median_values

def get_the_new_order(matrix):
    # Define the elements to check
    elements = np.array([1, 2, 3, 4])
    # Count the occurrences of each element along the rows
    counts = np.array([(matrix == e).sum(axis=1) for e in elements]).T
    # Sort elements by their counts in descending order
    sorted_indices = np.argsort(-counts, axis=1)
     # Get the sorted elements based on counts
    sorted_elements = np.take_along_axis(np.tile(elements, (matrix.shape[0], 1)), sorted_indices, axis=1)
    return sorted_elements

# Reorder the channel (ATCG to Max-min-3rd-4th)
def reorder_norm(combined_array, my_cmt):
    major_nt = my_cmt.major_nt.T
    order_base = get_the_new_order(
        major_nt)  # each row refers to one position, the 4 elements refer to the base index of "major", "minor", "other_1", "other_2"

    order_base -= 1
    reordered_array = np.take_along_axis(combined_array, order_base[:, np.newaxis, np.newaxis, :], axis=-1)
    ############ Order finished ##################
    first_two_rows = reordered_array[:, :, :2, :]
    sum_first_two = np.sum(first_two_rows, axis=(2, 3), keepdims=True)
    sum_first_two_fur = np.sum(sum_first_two, axis=1)
    exp_sum_first_two_fur = np.repeat(sum_first_two_fur, repeats=sum_first_two.shape[1], axis=1)
    exp_sum_first_two_fur = np.expand_dims(exp_sum_first_two_fur, axis=3)
    expanded_result = np.repeat(exp_sum_first_two_fur, repeats=4, axis=-1)
    last_row = np.max(reordered_array[:, :, -1:, :], axis=3, keepdims=True)
    expanded_last_row = np.repeat(last_row, repeats=4, axis=-1)
    normalized_first_two = reordered_array[:, :, :2, :] / expanded_result
    # Divide by the elements in the last row
    new_first_two = reordered_array[:, :, :2, :] / expanded_last_row
    new_first_two[new_first_two > 10] = 10
    normalized_first_two = np.nan_to_num(normalized_first_two, nan=0)
    new_first_two = np.nan_to_num(new_first_two, nan=0)
    new_array = reordered_array[:, :, :-1, :]
    final_array = np.concatenate([normalized_first_two, new_first_two, new_array], axis=2)
    return final_array

def find_sm_top_x(arr,x, mc):

    smallest_indices = np.argsort(arr.T, axis=1)[:, :x]
    # Create a boolean array initialized to False
    mask = np.zeros_like(arr.T, dtype=bool)
    # Use advanced indexing to set the smallest 3 elements to True
    np.put_along_axis(mask, smallest_indices, True, axis=1)
    mask=mask.T
    # Compute the median of each row
    medians = np.median(arr, axis=1)
    # Reshape the medians to align with the shape of the array
    medians = medians[:, np.newaxis]
    # Create a boolean mask for elements differing by more than 10 from the median
    mask2 = np.abs(arr - medians) > mc
    fmask=mask & mask2
    
    return fmask



def find_sm_top_x_test(arr, call_arr):
    #print(call_arr[40])
    #print(arr.T[40])
    #exit()
    #smallest_indices = np.argsort(arr.T, axis=1)[:, :3]
    masked_matrix = np.where(call_arr == 0, np.nan, call_arr)
    def most_frequent_nonzero(row):
        unique_elements, counts = np.unique(row[~np.isnan(row)], return_counts=True)
        if len(unique_elements) > 0:
            return unique_elements[np.argmax(counts)]
        else:
            return np.nan
    def least_frequent_nonzero(row):
        unique_elements, counts = np.unique(row[~np.isnan(row)], return_counts=True)
        if len(unique_elements) > 0:
            return unique_elements[np.argmin(counts)]
        else:
            return np.nan

    most_frequent_elements = np.apply_along_axis(most_frequent_nonzero, axis=1, arr=masked_matrix)
    least_frequent_elements = np.apply_along_axis(least_frequent_nonzero, axis=1, arr=masked_matrix)
    bool_max = call_arr == most_frequent_elements[:, np.newaxis]
    bool_min = call_arr == least_frequent_elements[:, np.newaxis]
    #print(arr.shape,bool_min.shape)
    #exit()
    c1 = arr.copy().T
    c2 = arr.copy().T
    #print(c1.shape,bool_min.shape)
    #exit()
    #print(c1[0], bool_max[0])
    #print(c2[0],bool_min[0])
    #exit()
    c1[~bool_max] = 0
    c2[~bool_min] = 0
    #print(c2[0])
    #exit()

    def median_no_zeros(row):
        non_zero_row = row[row != 0]  # remove 0
        if len(non_zero_row) == 0:  # if all 0, return 0
            return 0
        return np.median(non_zero_row)
    # Compute the median of each row
    #medians = np.apply_along_axis(median_no_zeros, axis=1, arr=c2)

    def mean_no_zeros(row):
        non_zero_row = row[row != 0]  # remove 0
        if len(non_zero_row) == 0:  # if all 0, return 0
            return 0
        return np.mean(non_zero_row)
    def std_no_zeros(row):
        non_zero_row = row[row != 0]  # remove 0
        if len(non_zero_row) == 0:  # if all 0, return 0
            return 0
        return np.std(non_zero_row)

    def iqr_no_zeros(row):
        non_zero_row = row[row != 0]  # 去掉0的元素
        if len(non_zero_row) == 0:  # 如果全是0，返回0
            return 0
        q75, q25 = np.percentile(non_zero_row, [75, 25])  # 计算75%分位数和25%分位数
        return 0.5*(q75 - q25)
    #mean_c1=np.apply_along_axis(mean_no_zeros,axis=1,arr=c1)
    median_c1 = np.apply_along_axis(median_no_zeros, axis=1, arr=c1)
    iqr_c1=np.apply_along_axis(iqr_no_zeros, axis=1, arr=arr)
    median_c1 = median_c1[:, np.newaxis]
    median_c1= np.tile(median_c1, (1, bool_min.shape[1]))
    #print(median_c1.shape,bool_min.shape)
    #print(median_c1[25])
    #exit()
    median_c1[~bool_min]=0
    #print(median_c1[25])
    #exit()
    iqr_c1=iqr_c1[:,np.newaxis]
    #print(median_c1.shape,c2.shape)
    res=median_c1-c2
    fmask = res > iqr_c1.T
    fmask=fmask.T

    return fmask

def cal_major_freq_in_call(arr1):
    #print(arr1.shape)
    #print(arr1)
    col_data_nonzero = [arr1[:, col][arr1[:, col] != 0] for col in range(arr1.shape[1])]
    '''
    for col in col_data_nonzero:
        print(col,np.bincount(col))
        print(np.argmax(np.bincount(col)))
    '''
    column_modes = [np.unique(col)[0] if len(np.unique(col)) == 1 else (1 if len(col) == 0 else np.argmax(np.bincount(col))) for col in col_data_nonzero]
    #exit()
    #print(column_modes)
    scount = np.sum(arr1 == column_modes, axis=0)
    #print(arr1==column_modes)
    #exit()
    return scount,arr1==column_modes



def compare_arrays_nonparametric(large_array_raw, small_array_raw):

    large_array=[x for x in large_array_raw if x != 0]
    #sorted_arr = np.sort(arr)
    #n = len(sorted_arr)
    #k = int(np.ceil(n * 0.25))
    #large_array = sorted_arr[:k]

    small_array=[x for x in small_array_raw if x != 0]
    #print(large_array,small_array)
    if len(small_array) == 0:
        return "Empty arrays! Please check."
        
    elif len(small_array) == 1:
        statistic, p_value = stats.wilcoxon(large_array - small_array[0])
    else:
        statistic, p_value = stats.mannwhitneyu(large_array, small_array)
    return p_value


def scan_continue_gap(arr):
    arr.sort()
    B = set()
    i, j = 0, 1
    n = len(arr)
    while i < n:
        while j < n and arr[j] - arr[i] <= 100:
            if arr[j] - arr[i] > 0:
                B.add(arr[i])
                B.add(arr[j])
            j += 1
        i += 1
        if j <= i:
            j = i + 1

    B = list(B)
    return B

def scan_continue_gap_revise(arr):
    arr.sort()
    #print(arr)
    B = {}

    i, j = 0, 1
    n = len(arr)
    #print(arr)
    while i < n:
        j=i+1
        #print('i,n:',i,',',n,'j,n:',j,',',n,'arr[j]-arr[i]:',arr[j],'-',arr[i])
        while j < n and arr[j] - arr[i] <= 100:
            #print(arr[j],arr[i])
            if arr[j] - arr[i] > 0:
                if arr[i] not in B:
                    B[arr[i]]={arr[j]:''}
                else:
                    B[arr[i]][arr[j]]=''
                if arr[j] not in B:
                    B[arr[j]]={arr[i]:''}
                else:
                    B[arr[j]][arr[i]]=''

            j += 1
            #print(i,j)
        i += 1
        if j <= i:
            j = i + 1
    res=[]
    #print(B)
    #print(B[1899148])
    '''
    arr_B=sorted(list(B.keys()))
    for key in B[arr_B[0]]:
        for key2 in B[arr_B[0]]:
            if not key==key2:
                B[key][key2]=''
    print(B)
    '''
    #print(check_arr)
    #exit()
    for p in B:
        if len(B[p])>2:
            res.append(p)
        '''
        else:
            check=0
            if len(B[p])==0:continue
            for s in B[p]:
                if s in check_arr:
                    check+=1
            if check==len(B[p]):
                res.append(p)
        # old code
        if len(B[p])>1:
            if len(B[p])>2:
                res.append(p)
            else:
                check=0
                for s in B[p]:
                    if s in check_arr:
                        check+=1
                if check==len(B[p]):
                    res.append(p)
        elif len(B[p])==1:
            check=1
            for e in B[p]:
                if e in check_arr and p in check_arr:
                    check=0
            if check==0:
                res.append(p)
        '''
    return res

def zscore_variant(large_array,small_array):
    mean_diff = np.mean(large_array) - np.mean(small_array)
    std_err_diff = np.std(large_array, ddof=1)
    z_score = mean_diff / std_err_diff
    #p_value = 2 * (1 - stats.norm.cdf(abs(z_score)))
    p_value=1 - stats.norm.cdf(abs(z_score))
    #print(mean_diff,z_score,std_err_diff)
    return z_score,p_value

def compare_arrays_ttest(large_array_raw, small_array_raw):
    #print(large_array,small_array)
    large_array=[x for x in large_array_raw if x != 0]
    #print('-> raw_array:',large_array)
    #mu=np.mean(large_array)
    #sigma = np.std(large_array)
    #alpha=0.01
    #threshold=norm.ppf(alpha, loc=mu, scale=sigma) 
    #print('-> threshold is ',threshold)
    large_array=np.array(large_array)
    #large_array_cdf=large_array[large_array<=threshold]
    #print('-> new_array:',large_array_cdf)
    '''
    sorted_arr = np.sort(arr)
    n = len(sorted_arr)
    k = int(np.ceil(n * 0.25))
    large_array = sorted_arr[:k]
    '''

    small_array=[x for x in small_array_raw if x != 0]
    #print(large_array,small_array)
    #exit()
    n_small = len(small_array)
    normc=1
    if n_small == 0:
        return "Empty array! Please check."
        exit()
    
    elif n_small == 1:
        target_value = small_array[0]
        t_stat, p_value = stats.ttest_1samp(large_array, popmean=target_value)
        # new p-cdf with z-score variant
        z_score,p_cdf=zscore_variant(large_array,small_array)
        '''
        if not (len(large_array_cdf)==0 or len(large_array_cdf)==1):
            t_stat, p_cdf = stats.ttest_1samp(large_array_cdf, popmean=target_value)
        else:
            normc=0
            p_cdf=p_value
        '''
        #print('-> 2 p-values: ',p_value,p_cdf)
        if np.isnan(p_cdf):
            normc=0
        return p_value,p_cdf
    
    else:
        t_stat, p_value = stats.ttest_ind(large_array, small_array, equal_var=False)
        z_score,p_cdf=zscore_variant(large_array,small_array)
        '''
        if not (len(large_array_cdf)==0 or len(large_array_cdf)==1):
            t_stat, p_cdf = stats.ttest_ind(large_array_cdf,small_array,equal_var=False)
        else:
            normc=0
            p_cdf=p_value
        '''
        #print('-> 2 p-values: ',p_value,p_cdf)
        if np.isnan(p_cdf):
            normc=0
        return p_value,p_cdf

def check_mm(arr1,counts_major):
    #print(arr1[:,:10])
    col_data_nonzero = [arr1[:, col][arr1[:, col] != 0] for col in range(arr1.shape[1])]
    #print(col_data_nonzero)
    #exit()
    '''
    column_modes=[]
    for col in col_data_nonzero:
        #print(col,np.bincount(col),np.unique(col))
        if len(np.unique(col)) >= 2
            nonzero_pairs = sorted([(x, i) for i, x in enumerate(np.bincount(col)) if x != 0], reverse=True)
            minor=nonzero_pairs[1][1] if len(nonzero_pairs) > 1 else -1
            
        else:
            if len(col) == 0:
                minor=1
            else:
                minor=0
        column_modes.append(minor)
        #print(res)
        exit()
    '''
    column_modes = [sorted([(x, i) for i, x in enumerate(np.bincount(col)) if x != 0], reverse=True)[1][1] if len(np.unique(col)) >= 2 and len([(x, i) for i, x in enumerate(np.bincount(col)) if x != 0]) > 1 else (1 if len(col) == 0 else 0) for col in col_data_nonzero]
    #print(column_modes)
    #exit()
    #column_modes = [np.unique(col)[1] if len(np.unique(col)) >= 2 else (1 if len(col) == 0 else 0) for col in col_data_nonzero]
    scount = np.sum(arr1 == column_modes, axis=0)
    only_minor= arr1 == column_modes
    #print(only_minor.shape,counts_major.shape)
    #exit()
    #print(scount)
    check_minor_more_than_1=np.repeat([scount>1], arr1.shape[0], axis=0)
    cmj_copy=copy.deepcopy(counts_major)
    cmj_copy[~only_minor]=0
    #print(counts_major)
    min_v=[min(col[col != 0]) if len(col[col != 0]) > 0 else 0 for col in cmj_copy.T]
    #print(min_v)
    #exit()
    c,b=cal_major_freq_in_call(arr1)
    copy_minv=np.repeat([min_v],arr1.shape[0],axis=0)
    #print('cminv:',copy_minv)
    #print('cmajor:',counts_major)
    x=counts_major-copy_minv
    x[only_minor]=-1
    #print('cmj-cmi:',x)

    #print(counts_major,arr1)
    #print(arr1.shape,check_minor_more_than_1.shape)
    #exit()
    ##print(column_modes)
    #print(scount)
    #print(arr1)
    #c,b=cal_major_freq_in_call(arr1)
    #minor_max = check_minor_more_than_1
    minor_max = check_minor_more_than_1 & (x<0)
    #exit()
    #print(minor_max)
    #print(arr1)
    #exit()
    
    return minor_max,arr1==column_modes

def process_arrays(arr1, arr2, sample_num):
    col_data_nonzero = [arr1[:, col][arr1[:, col] != 0] for col in range(arr1.shape[1])]
    column_modes = [np.unique(col)[0] if len(np.unique(col)) == 1 else ( 1 if len(col)==0 else np.argmax(np.bincount(col))) for col in col_data_nonzero]
    #print(arr2)
    scount = np.sum(arr1 == column_modes, axis=0)
    #print(scount)
    mask = arr1 != np.array(column_modes)
    arr2[mask] = 0
    result = np.sum((arr2 > 0) & (arr2 < 1), axis=0)
    #print(result/scount)
    #exit()

    return result/scount

def cal_freq_amb_samples(all_p,my_cmt):
    #print(len(all_p))
    #print(len(my_cmt.p))
    #exit()
    keep_col=[]
    for p in my_cmt.p:
        if p in all_p:
            keep_col.append(True)
        else:
            keep_col.append(False)
    keep_col=np.array(keep_col)
    #print(len(my_cmt.p))
    #print(len(keep_col))
    #print(keep_col)
    #exit()
    my_cmt.filter_positions(keep_col)
    #exit()
    freq_arr=process_arrays(my_cmt.major_nt,my_cmt.major_nt_freq,my_cmt.major_nt.shape[0])
    ##exit()
    freq_d={}
    c=0
    for p in my_cmt.p:
        freq_d[p]=freq_arr[c]
        c+=1
    return freq_d,freq_arr

def remove_lp(combined_array,inp,my_cmt,my_calls, median_cov ):
    raw_p=len(inp)
    rawp=inp
    #print(inp.shape,combined_array[40])
    #exit()
    ######### Further Scan bad pos, eg: potential FPs caused by Low-Depth samples
    #print(my_cmt.p)
    keep_col = []
    for pos in my_cmt.p:
        if pos not in inp:
            keep_col.append(False)
        else:
            keep_col.append(True)
    keep_col = np.array(keep_col)
    #print(keep_col.shape)
    #print(my_cmt.p.shape)
    #print(my_calls.p.shape)
    #exit()

    my_cmt.filter_positions(keep_col)
    my_calls.filter_positions(keep_col)
    #print(my_cmt.p)
    #exit()
    my_calls.filter_calls_by_element(
        my_cmt.fwd_cov < 1
    )

    my_calls.filter_calls_by_element(
        my_cmt.rev_cov < 1
    )
    my_calls_check=copy.deepcopy(my_calls)
    scount_before,b=cal_major_freq_in_call(my_calls_check.calls)
    #print(scount_before[:100])
    #exit()
    my_calls_check.filter_calls_by_element(
            (my_cmt.fwd_cov < 11) & (my_cmt.rev_cov < 11)
    )
    scount_after,b=cal_major_freq_in_call(my_calls_check.calls)
    #print(scount_after[:100])
    fratio=scount_after/scount_before
    #print(fratio)
    count=np.sum(fratio > 0.5)
    #print(count)
    stat=count/len(fratio)
    #print(stat)
    #exit()
    my_cmt_tem=copy.deepcopy(my_cmt)
    freq_p,freq_arr=cal_freq_amb_samples(my_calls.p,my_cmt_tem)
    freq_check=np.repeat([freq_arr>0],my_calls.calls.shape[0],axis=0) 
    if count>10:
        my_calls.filter_calls_by_element(
            (my_cmt.fwd_cov < 3) & (my_cmt.rev_cov < 3) & freq_check
        )
    else:
        if stat>=0.1:
            my_calls.filter_calls_by_element(
                    (my_cmt.fwd_cov < 3) & (my_cmt.rev_cov < 3) & freq_check
            )
    # if fwd type != rev type -> filter
    #print(np.where(my_cmt.p==1161288))
    my_calls.filter_calls_by_element(
            (my_cmt.major_nt_fwd != my_cmt.major_nt_rev) & ((my_cmt.major_nt_fwd>0)&(my_cmt.major_nt_rev>0))
    )
    '''
    # new rule - large fwd/rev and very samll rev/fwd - >70% sample fwd>4*rev or rev>4*fwd, and fwd or rev <10
    my_calls_check=copy.deepcopy(my_calls)
    cbefore=np.sum(my_calls_check.calls != 0, axis=0) 
    my_calls_check.filter_calls_by_element(
            ((my_cmt.fwd_cov - 4*my_cmt.rev_cov >0 ) | (my_cmt.rev_cov - 4*my_cmt.fwd_cov > 0)) & ((my_cmt.fwd_cov < 10) | (my_cmt.rev_cov < 10))
            )
    cafter=np.sum(my_calls_check.calls != 0, axis=0)
    
    diff_ratio=1-cafter/cbefore
    keep_col=[]
    if my_calls_check.calls.shape[0]>10:
        cut=0.6
    else:
        cut=0.7
    for d in diff_ratio:
        if d>cut:
            keep_col.append(False)
        else:
            keep_col.append(True)
    keep_col=np.array(keep_col)
    my_cmt.filter_positions(keep_col)
    my_calls.filter_positions(keep_col)
    '''

    #print(diff_ratio,my_calls_check.p)
    #exit()

    my_calls.filter_calls_by_element(
        my_cmt.major_nt_freq < 0.7
    )
    #print(my_cmt.p[:10])
    # print(my_cmt.counts_major.shape)
    # print(my_cmt.counts_minor.shape)
    # na=my_cmt.counts_major-5*my_cmt.counts_minor
    # test=my_cmt.major_nt_freq < 0.7
    # t2= my_cmt.counts_minor>20
    #a=np.array(my_cmt.counts_major, dtype=np.int64)
    #b=np.array(my_cmt.counts_minor, dtype=np.int64)
    #print(my_calls.calls[:,np.where(my_cmt.p==1052499)])
    # exit()
    ### Super big fp pos filter - 2024-12-18
    #rawp=my_c.p
    #my_calls.filter_calls_by_element(
    #    (my_cmt.counts_minor>30) & ((a-b*3)>0) & (my_cmt.counts_major>250)
    #)
    #print(my_cmt.counts_major[:,np.where(my_cmt.p==1052499)]-(my_cmt.counts_minor[:,np.where(my_cmt.p==1052499)]*10))
    #print((a-b*10)[:,np.where(my_cmt.p==1052499)])
    #exit()
    #print('filter_low_quality_pos...')
    #exit()
    if np.median(median_cov)>9:
        my_calls.filter_calls_by_element(
            ((my_cmt.rev_cov == 1 ) & (my_cmt.fwd_cov==1)) | ((my_cmt.rev_cov == 1 ) & (my_cmt.fwd_cov<4)) | ((my_cmt.rev_cov < 4 ) & (my_cmt.fwd_cov==1))
        )

    #print(combined_array.shape)
    #exit()

    if np.median(median_cov) > 20 and combined_array.shape[1]>50:
        #print(median_cov,median_cov.shape)
        #exit()
        my_calls.filter_calls_by_element(
            my_cmt.fwd_cov < 4
        )

        my_calls.filter_calls_by_element(
            my_cmt.rev_cov < 4
        )




    keep_col = remove_same(my_calls)
    #print(keep_col)
    #exit()
    my_cmt.filter_positions(keep_col)
    my_calls.filter_positions(keep_col)

    keep_col_arr=[]
    for s in inp:
        if s in my_calls.p:
            keep_col_arr.append(True)
        else:
            keep_col_arr.append(False)
    keep_col_arr=np.array(keep_col_arr)
    inp=inp[keep_col_arr]
    combined_array=combined_array[keep_col_arr]
    #print(my_cmt.p)
    #exit()
    #### Filter low gap pos
    #print('before-filt-gap:',my_cmt.p)
    rawp=my_cmt.p # used to check positions removed by gap filter
    #exit()
    '''
    # new rule - large fwd/rev and very samll rev/fwd - >70% sample fwd>4*rev or rev>4*fwd, and fwd or rev <10  & freq must >0
    my_calls_check=copy.deepcopy(my_calls)
    if my_calls_check.calls.shape[0]>10:
        cut=0
    else:
        cut=0.2
    freq_p,freq_arr=cal_freq_amb_samples(my_calls_check.p,my_cmt)
    freq_check=np.repeat([freq_arr>cut],my_calls_check.calls.shape[0],axis=0) 

    cbefore=np.sum(my_calls_check.calls != 0, axis=0)
    my_calls_check.filter_calls_by_element(
            ((my_cmt.fwd_cov - 4*my_cmt.rev_cov >0 ) | (my_cmt.rev_cov - 4*my_cmt.fwd_cov > 0)) & ((my_cmt.fwd_cov < 10) | (my_cmt.rev_cov < 10)) & freq_check
            )
    cafter=np.sum(my_calls_check.calls != 0, axis=0)

    diff_ratio=1-cafter/cbefore
    keep_col=[]
    if my_calls_check.calls.shape[0]>10:
        cut=0.6
    else:
        cut=0.7
    for d in diff_ratio:
        if d>cut:
            keep_col.append(False)
        else:
            keep_col.append(True)
    keep_col=np.array(keep_col)
    my_cmt.filter_positions(keep_col)
    my_calls.filter_positions(keep_col)
    ########################### new rule done
    '''
    med_cov_ratio_fwd=my_cmt.fwd_cov/median_cov[:, np.newaxis]
    med_cov_ratio_rev = my_cmt.rev_cov / median_cov[:, np.newaxis]
    med_cov_ratio_fwd = np.nan_to_num(med_cov_ratio_fwd, nan=0)
    med_cov_ratio_rev = np.nan_to_num(med_cov_ratio_rev, nan=0)
    # Add super big fp filt in gap filt
    a=np.array(my_cmt.counts_major, dtype=np.int64)
    b=np.array(my_cmt.counts_minor, dtype=np.int64)
    #print(a.shape)
    #print(my_calls.calls[:,np.where(my_cmt.p==1052499)])
    # exit()
    #rawp=my_c.p
    #print(my_calls.p)
    minor_max,mbool=check_mm(my_calls.calls,a) # Whether the major  counts all < the min minor count & # of minor > 1
    my_calls.filter_calls_by_element(
        (my_cmt.counts_minor>30) & ((a-b*3)>0) & (my_cmt.counts_major>250) & minor_max
    )
    #print(my_cmt.fwd_cov)
    #print((my_cmt.counts_minor>30) & ((a-b*3)>0) & (my_cmt.counts_major>250))
    #print(((my_cmt.fwd_cov>200) | (my_cmt.rev_cov>200)))

    #my_calls.filter_calls_by_element(
    #    (my_cmt.counts_minor>30) & ((a-b*3)>0) & (my_cmt.counts_major>250)
    #)
    #print(my_cmt.p)
    #print(my_cmt.fwd_cov.shape)
    #print(my_cmt.fwd_cov)
    #print(my_cmt.fwd_cov<200)
    #print(my_calls.p)
    #print(combined_array[40][21])
    #print(med_cov_ratio_fwd[40][21])
    #print(med_cov_ratio_rev[40][21])
    #print(combined_array.shape,my_cmt.fwd_cov.shape)
    #exit()
    # Old methods: Use loose filtering
    #mask1 = find_sm_top_x(my_cmt.fwd_cov, 3, 20)
    #mask2 = find_sm_top_x(my_cmt.rev_cov, 3, 20)
    # old gap rule  - 2025-02-03
    '''
    if combined_array.shape[1]<25:
        mask1=find_sm_top_x(my_cmt.fwd_cov, 3, 20)
        mask2=find_sm_top_x(my_cmt.rev_cov, 3, 20)
    else:
        mask1 = find_sm_top_x(my_cmt.fwd_cov, 5, 20)
        mask2 = find_sm_top_x(my_cmt.rev_cov, 5, 20)
    '''

    ####### old done #########################
    #print(mask1.shape)
    #print(my_calls.calls.shape)
    #exit()
    #print(np.where(my_cmt.p==864972))
    #exit()
    # New methods: Test IQR-based filtering
    #mask1 = find_sm_top_x_test(my_cmt.fwd_cov,my_calls.calls.T)
    #mask2 = find_sm_top_x_test(my_cmt.rev_cov,my_calls.calls.T)
    #exit()
    #print(med_cov_ratio_fwd[:,25])
    ##print(med_cov_ratio_fwd)
    #print(mask1[:,25])
    #exit()

    # old gap rule - 2025-02-03
    '''
    mask1 = mask1 & (med_cov_ratio_fwd< 0.1)
    mask2 = mask2 & (med_cov_ratio_rev < 0.1)
    mask=mask1 & mask2
    my_calls.filter_calls_by_element(
        mask
    )
    '''


    # Add gap rule: if all minor sample <=10 and one part must <=5, >=85% major sample not satisfy this rule and >=20, then should be gap fp
    my_calls_tem=copy.deepcopy(my_calls)
    #print(1899148 in my_calls_tem.p)
    rawp_tem=my_calls_tem.p
    scount_before,b=cal_major_freq_in_call(my_calls_tem.calls)
    #count_combine=my_cmt.counts_major
    count_combine_ratio=my_cmt.counts_major/median_cov[:, np.newaxis]
    #print(my_cmt.counts_major,my_cmt.counts_major[:,temp])
    #print(temp)
    #exit()
    #print(temp,my_cmt.counts_major,my_cmt.counts_major.shape,b.shape,mbool.shape)
    #exit()
    #c2=copy.deepcopy(count_combine)
    c3=copy.deepcopy(count_combine_ratio)
    #count_combine[~b]=0
    count_combine_ratio[~b]=0
    #c2[~mbool]=0
    c3[~mbool]=0
    p_arr=[]
    p_arr_ratio=[]
    p_arr_ratio_cdf=[]
    #normc=[]
    tem=[]
    tem2=[]
    #p_non=[]
    for i in range(count_combine_ratio.shape[1]):
        #a_arr=count_combine[:,i]
        #b_arr=c2[:,i]
        c_arr=count_combine_ratio[:,i]
        d_arr=c3[:,i]
        #a1=[x for x in a_arr if x != 0]
        #b1=[x for x in b_arr if x != 0]
        c1=[x for x in c_arr if x != 0]
        d1=[x for x in d_arr if x != 0]
        tem.append([c1,d1])
        #tem2.append([a1,b1])
        #p2=compare_arrays_ttest(a_arr,b_arr) # Count
        p,p_cdf=compare_arrays_ttest(c_arr,d_arr) # Ratio
        #p,effect_size, power=compare_groups(c_arr, d_arr)
        #p2=compare_arrays_nonparametric(c_arr,d_arr) # Ratio
        #p_non.append(p2)
        #p2=compare_arrays_nonparametric(a_arr,b_arr) # Count
        #p_arr.append(p2)
        p_arr_ratio.append(p)
        #tem2.append([effect_size,power])
        p_arr_ratio_cdf.append(p_cdf)
        #normc.append(norm_check)
    #print(p_arr,p_arr_ratio)
    gap_candidate=[]
    tem_p=[]
    #tpos=[3950318,2403312,78143,960515,1011874,624313,4129099]
    #label=[0,0,0,1,1,1,1]
    #dtl=dict(zip(tpos,label))
    #o=open('position_vector_to_Tami/p11.txt','w+')
    #for p in range(len(p_arr_ratio))
    ct=0
    for i in range(len(p_arr_ratio)):
        '''
        if my_cmt.p[i] in tpos:
            o.write('positions\tnorm_depth_major_isolates\tnorm_depth_minor_isolates\tlabel\n')
            o.write(str(my_cmt.p[i])+'\t'+','.join(map(str, tem[i][0]))+'\t'+','.join(map(str, tem[i][1]))+'\t'+str(dtl[my_cmt.p[i]])+'\n')
        '''   

        if p_arr_ratio_cdf[i] <0.01:
            tem[i][0]=np.array(tem[i][0])
            if max(tem[i][1])<min(tem[i][0]) and max(tem[i][1])<0.05:
                #raw: if max(tem[i][0])>0.1 and len(tem[i][0][tem[i][0]<0.2])/len(tem[i][0])<0.5:
                if max(tem[i][0])>0.2:
                    gap_candidate.append(my_cmt.p[i])
                '''
                # old rule
                if normc[i]==1:
                    #if len(tem[i][0][tem[i][0]<0.1])/len(tem[i][0])<0.5:
                    if max(tem[i][0])>0.1:
                        gap_candidate.append(my_cmt.p[i])
                else:
                    #print('')
                    #if len(tem[i][0][tem[i][0]<0.1])/len(tem[i][0])<0.3:
                    #print(tem[i][0])
                    if len(tem[i][0][tem[i][0]>0.1])/len(tem[i][0])>=0.7:
                        gap_candidate.append(my_cmt.p[i])
                '''
            '''
            elif min(tem[i][1])>max(tem[i][0]) and min(tem[i][1])>0.05 and max(tem[i][0])<0.05:
                gap_candidate.append(my_cmt.p[i])
            '''
        if p_arr_ratio_cdf[i] <0.01 or p_arr_ratio[i] <0.05:
            if np.max(tem[i][1])<0.1 and len(tem[i][1])<=3 and max(tem[i][0])>0.2:
                tem_p.append(my_cmt.p[i])
            tem[i][0]=np.array(tem[i][0])
            #print('pos:',my_cmt.p[i],'\n','p-value:',p_arr_ratio[i],'\np-value-z-score:',p_arr_ratio_cdf[i],'\nnorm_major_mean:',np.mean(tem[i][0]),'\nnorm_major_median:',np.median(tem[i][0]),'\n# of major <0.2',np.sum(tem[i][0]<0.2),'\nthese elements are:',tem[i][0][tem[i][0]<0.2],'\nnorm_minor_arr:',tem[i][1])
        #print(my_cmt.p[i],tem[i][1])
        '''
        if p_arr_ratio[i]<0.05 and np.max(tem[i][1])<0.1 :
            p=1
            #print(my_cmt.p[i],p_arr_ratio[i],p_arr[i],tem[i][0],tem[i][1])
            gap_candidate.append(my_cmt.p[i])
            #gap_pos.append()
            tem[i][0]=np.array(tem[i][0])
            # old one with # '\nnorm_major_arr:',tem[i][0]
            print('pos:',my_cmt.p[i],'\n','p-value:',p_arr_ratio[i],'\nnorm_major_mean:',np.mean(tem[i][0]),'\nnorm_major_median:',np.median(tem[i][0]),'\n# of major <0.1',np.sum(tem[i][0]<0.1),'\nthese elements are:',tem[i][0][tem[i][0]<0.1],'\nnorm_minor_arr:',tem[i][1])
            if np.sum(tem[i][0]<0.1)>1:
                if np.min(tem[i][0])<np.max(tem[i][1]) or np.min(tem[i][0])-np.max(tem[i][1])<0.01 or np.sum(tem[i][0]<0.1)>0.2*len(tem[i][0]):
                    print('\n')
                else:
                    p=compare_arrays_ttest(tem[i][0][tem[i][0]<0.1],tem[i][1]) 
                    print('\np-value-major-s0.1-minor:',p)
                    print('\nnorm_major_s0.1_median:',np.median(tem[i][0][tem[i][0]<0.1]))
                    if p<0.05:
                        tem_p.append(my_cmt.p[i])
            else:
                if np.sum(tem[i][0]<0.1)==1:
                    #print(tem[i][0][tem[i][0]<0.1][0])
                    if  np.min(tem[i][1])>0.05:
                        print('\n')
                    elif not tem[i][0][tem[i][0]<0.1][0]<0.09:
                        tem_p.append(my_cmt.p[i])
                    elif np.max(tem[i][1])<0.02 and np.min(tem[i][0])-np.max(tem[i][1])>0.05:
                        tem_p.append(my_cmt.p[i])
                else:
                    if np.min(tem[i][1])<0.05:
                        tem_p.append(my_cmt.p[i])
            #if np.sum(tem[i][0]<0.1)==0 or (np.sum(tem[i][0]<0.1)==1 and tem[i][0][0]>0.05) or (p<0.05 and np.median(tem[i][1])<np.median(tem[i][0][tem[i][0]<0.1])):
                #tem_p.append(my_cmt.p[i])
        
            #print(p_arr[i],tem2[i][0],tem2[i][1])
        '''
    #print(gap_candidate)
    #exit()
    gap_pos_add=scan_continue_gap_revise(tem_p)
    #gap_pos_add=[]
    gap_pos=gap_candidate
    
    for p in gap_pos_add:
        if p not in gap_pos:
            gap_pos.append(p)
    
    print('gap_pos:\n',gap_pos)
    #print(p_arr[36],p_arr_ratio[36])
    #exit()
    #print(my_cmt.p[0],my_cmt.p[36])

    #exit()
    #print(scount_before)
    #print('before:',my_calls_tem.calls)
    #exit()
    # old gap rule - 2025-02-03
    '''
    my_calls_tem.filter_calls_by_element(
        ((my_cmt.fwd_cov <= 10) & (my_cmt.rev_cov <= 10)) & ((my_cmt.fwd_cov <= 5) | (my_cmt.rev_cov <= 5))
    )

    #print(my_cmt.major_nt,my_cmt.major_nt.shape)
    #print(my_cmt.fwd_cov,my_cmt.fwd_cov.shape)
    #exit()

    my_calls_tem.filter_calls_by_element(
            (my_cmt.fwd_cov <= 20) & (my_cmt.rev_cov <= 20) & b
    )
    
    scount_after,b=cal_major_freq_in_call(my_calls_tem.calls)
    keep_col_tem = remove_same(my_calls_tem)
    #print(keep_col_tem)
    if len(keep_col_tem)==0:
        print('No SNPs detected! Exit.')
        exit()
    #exit()
    my_calls_tem.filter_positions(keep_col_tem)
    gap_ratio=scount_after/scount_before
    dtem=dict(zip(rawp_tem,gap_ratio))
    #print(dtem)
    #print(scount_after)
    #print('after:',my_calls_tem.calls)
    #print(my_calls_tem.p,gap_ratio)
    gap_pos=[]
    for p in rawp_tem:
        if p not in my_calls_tem.p:
            if dtem[p]>=0.85:
                gap_pos.append(p)
    #exit()
    #print(gap_pos)
    #exit()
    

    '''
    #print(my_calls.p, len(my_calls.p))
    keep_col_raw = remove_same(my_calls)
    keep_col=[]
    c=0
    for k in keep_col_raw:
        if my_calls.p[c] in gap_pos:
            keep_col.append(False)
        else:
            keep_col.append(k)
        c+=1
    keep_col=np.array(keep_col)
    my_cmt.filter_positions(keep_col)
    my_calls.filter_positions(keep_col)
    #print('after-filt-gap:',my_cmt.p)
    pfgap=[p for p in rawp if p not in my_cmt.p] # position filterd due to gap
    #print(len(pfgap))
    #exit()
    #print(my_calls.p, len(my_calls.p))
    #print(pfgap)
    keep_col_arr = []
    for s in inp:
        if s in my_calls.p:
            keep_col_arr.append(True)
        else:
            keep_col_arr.append(False)
    keep_col_arr = np.array(keep_col_arr)
    inp = inp[keep_col_arr]
    combined_array = combined_array[keep_col_arr]

    print('There are ',raw_p-len(inp),' pos filtered. Keep ',len(inp),' positions.')
    print([p for p in rawp if p not in inp])
    #exit()
    return combined_array,inp,pfgap,rawp


# filter low-quality samples and low-quality positions
def remove_low_quality_samples(inarray,thred,inpos):
    #print(inpos)
    #print(inarray.shape)
    trans = np.transpose(inarray, (1, 0, 2, 3))
    raw_sample=trans.shape[0]
    sum1=np.sum(trans,axis=3)
    sum2=np.sum(sum1,axis=2)
    # Count the number of zeros in each row
    zero_count = np.sum(sum2 == 0, axis=1)

    # Calculate the percentage of zeros
    percent_zeros = (zero_count / sum2.shape[1]) * 100
    trans=trans[percent_zeros<thred]
    sum1_new = np.sum(trans, axis=3)
    new_sample=trans.shape[0]
    trans = np.transpose(trans, (1, 0, 2, 3))
    print('Remove ',raw_sample-new_sample,' low-quality samples!')
    raw_pos=trans.shape[0]
    ########
    tem = trans[:, :, 4:6, :]
    #print(tem[6,:])
    ########## filter the position with the same type of base
    if raw_sample-new_sample>0:
        check=tem[:,:,:,1]==0
        check=np.sum(check,axis=1)
        check = np.sum(check, axis=1)
        check_2 = tem[:, :, :, 0] != 0
        # check 0-lines
        all_zero=tem==0
        az=np.sum(all_zero,axis=-1)
        zl=az==4
        find_minor=tem[:, :, :, 1] - tem[:, :, :, 0]
        find_minor=find_minor>0
        find_minor=~find_minor
        check_2=(check_2 | zl ) & find_minor

        check_2 = np.sum(check_2, axis=1)
        check_2 = np.sum(check_2, axis=1)
        keep_col_1=check!=trans.shape[1]*2

        keep_col_2=check_2==trans.shape[1]*2
        keep_col_2=~keep_col_2

        keep_col=keep_col_1 & keep_col_2

        trans=trans[keep_col]
        inpos=inpos[keep_col]
        
        unqiue_pos=trans.shape[0]
        print('Remove ',raw_pos-unqiue_pos,' same positions!')
        #print(inpos_out)
        # c=0
        # t=0
        # for i in inpos:
        #     if i not in inpos_out and inlab[c]==1:
        #         print(i)
        #         t+=1
        #         if t>9:
        #             exit()
        #     c+=1

        #print(np.where(inpos_out==676632))
        #exit()
        #exit()
        #else:
        #inpos_out=inpos
        #outlab=inlab

    ##########  filter low-quality positions
    ## - new try
    tem=trans[:, :, 4:6, :]

    ###### C31 refers to the minor sample
    c1=trans[:,:,0,1]>0
    c2=trans[:,:,1,1]>0
    c31=c1 | c2
    c31_b= c1 & c2 # pure minor

    ###### C32 -> Check whether both rev and fwd >0
    c1 = np.sum(tem[:,:,0,:],axis=-1)>0
    c2 = np.sum(tem[:, :, 1, :],axis=-1)>0
    c32= c1 & c2

    ###### C33 -> Both fwd and rev are pure (only 1 type of non-zero base)
    c3=np.sum(tem>0,axis=-1)==1
    c33=np.sum(c3,axis=-1)==2

    ###### C34 -> fwd and rew has different bases
    c1=np.argmax(tem[:,:,0,:],axis=-1)
    c2 = np.argmax(tem[:, :, 1, :], axis=-1)
    c34=c1!=c2
    call=c31 & c32 & c33 & c34
    pure_minor = c31_b & c32 & c33
    fc1=np.sum(call,axis=-1)>0
    fc2 = np.sum(pure_minor, axis=-1)<2
    ### Stat how many pure minor samples
    keep_col=fc1 & fc2
    keep_col=~keep_col
    raw_pos=len(keep_col)
    trans=trans[keep_col]
    inpos=inpos[keep_col]
    #inpos_out = inpos_out[keep_col]

    #outlab = outlab[keep_col]
    new_pos=trans.shape[0]
    print('Remove ',raw_pos-new_pos,' low-quality positions! Finally remains ',new_pos,' positions!')

    return trans,inpos

def trans_shape(indata):
    return np.transpose(indata, (1, 0, 2, 3))

def remove_same(my_calls_in):
	keep_col = []
	for i in range(my_calls_in.calls.shape[1]):
		unique_nonzero_elements = np.unique(my_calls_in.calls[:, i][my_calls_in.calls[:, i] != 0])
		if len(unique_nonzero_elements) < 2:
			my_calls_in.calls[:, i] = 0
			keep_col.append(False)
		else:
			keep_col.append(True)
	keep_col = np.array(keep_col)
	return keep_col

def load_p(infile):
    f=open(infile,'r')
    line=f.readline()
    d={}
    while True:
        line=f.readline().strip()
        if not line:break
        ele=line.split('\t')
        d[int(ele[0])]=''
    return d

def remove_exclude_samples(samples_to_exclude_bool,quals,counts,sample_names,indel_counter,raw_cov_mat,in_outgroup):
    in_outgroup=samples_to_exclude_bool
    quals = quals[~in_outgroup]
    counts = counts[~in_outgroup]
    sample_names = sample_names[~in_outgroup]
    indel_counter = indel_counter[~in_outgroup]
    raw_cov_mat = raw_cov_mat[~in_outgroup]
    in_outgroup=in_outgroup[~in_outgroup]
    return quals,counts,sample_names,indel_counter,raw_cov_mat,in_outgroup


def data_transform(infile,incov,fig_odir,samples_to_exclude,min_cov_samp):

    # infile='../../../Scan_FP_TP_for_CNN/Cae_files/npz_files/Lineage_10c/candidate_mutation_table_cae_Lineage_10c.npz'
    [quals, p, counts, in_outgroup, sample_names, indel_counter] = \
        snv.read_candidate_mutation_table_npz(infile)

    #print(in_outgroup,in_outgroup.shape)
    if not len(in_outgroup)==len(sample_names):
        in_outgroup=np.array([False] * len(sample_names))
    # Only for P15
    #in_outgroup[2]=True
    #in_outgroup[13]=True
    #in_outgroup[13]=True

    ########  remove outgroup samples
    quals = quals[~in_outgroup]
    counts = counts[~in_outgroup]
    sample_names = sample_names[~in_outgroup]
    indel_counter = indel_counter[~in_outgroup]
    raw_cov_mat = snv.read_cov_mat_npz(incov)
    raw_cov_mat = raw_cov_mat[~in_outgroup]
    in_outgroup=in_outgroup[~in_outgroup]
    #in_outgroup=np.array([False] * len(sample_names))
    #in_outgroup[2]=True
    #in_outgroup[13]=True
    
    my_cmt = snv.cmt_data_object(sample_names,
                                 in_outgroup,
                                 p,
                                 counts,
                                 quals,
                                 indel_counter
                                 )
    samples_to_exclude_bool = np.array( [x in samples_to_exclude for x in my_cmt.sample_names] )
    my_cmt.filter_samples( ~samples_to_exclude_bool )
    quals,counts,sample_names,indel_counter,raw_cov_mat,in_outgroup=remove_exclude_samples(samples_to_exclude_bool,quals,counts,sample_names,indel_counter,raw_cov_mat,in_outgroup)

    my_calls = snv.calls_object(my_cmt)

    keep_col = remove_same(my_calls)

    my_cmt.filter_positions(keep_col)
    my_calls.filter_positions(keep_col)
    
    quals = quals[:,keep_col]
    counts = counts[:,keep_col,:]
    p=p[keep_col]
    #sample_names = sample_names[in_outgroup]
    indel_counter=indel_counter[:,keep_col,:]
    median_cov = np.median(raw_cov_mat, axis=1)
    #print(median_cov.shape)
    #exit()
 
    ###### Single pos fig check

    #check=0
    '''
    for i in range(len(p)):
        # only for checking 13b
        #widp=load_p('../../../../../../Downloads/clabsi_project/analysis/output_new/Klebsiella_pneumoniae_P-13_b/snv_table_mutations_annotations.tsv')
        #if p[i] not in widp:continue
        #check+=1
        # done for 13 b
        snv.single_point_plot_herui_add_median(my_cmt.counts, i, str(p[i]), my_cmt.sample_names, 'Unknown-'+fig_odir , 'Col-pos-check-single/'+fig_odir, median_cov)
        snv.single_point_plot_herui(my_cmt.counts, i, str(p[i]), my_cmt.sample_names, 'Unknown-' + fig_odir,'Col-pos-check-single/' + fig_odir+'_no_median')
        #exit()
    #print(check)
    '''

    indata_32 = counts
    indel = indel_counter
    qual = quals
    indel= np.sum(indel, axis=-1)

    expanded_array = np.repeat(indel[:, :, np.newaxis], 4, axis=2)
    expanded_array_2 = np.repeat(qual[:, :, np.newaxis], 4, axis=2)
    med_ext = np.repeat(median_cov[:, np.newaxis], 4, axis=1)
    med_arr = np.tile(med_ext, (counts.shape[1], 1, 1))

    new_data = indata_32.reshape(indata_32.shape[0], indata_32.shape[1], 2, 4)
    new_data=trans_shape(new_data)
    
    indel_arr_final = np.expand_dims(expanded_array, axis=2)
    indel_arr_final=trans_shape(indel_arr_final)
    qual_arr_final = np.expand_dims(expanded_array_2, axis=2)
    qual_arr_final=trans_shape(qual_arr_final)
    med_arr_final = np.expand_dims(med_arr, axis=2)
    combined_array = np.concatenate((new_data, qual_arr_final, indel_arr_final, med_arr_final), axis=2)
    check_idx = 0
    c1 = (combined_array[..., :] == 0)
    x1 = (np.sum(c1[:, :, :2, :], axis=-2) == 2)
    mx = x1
    mxe = np.repeat(mx[:, :, np.newaxis, :], 5, axis=2)
    combined_array[mxe] = 0
    ####### Reorder the columns and normalize & split the count info
    '''
    keep_col = []
    # print(my_cmt.p)
    # print(inpos)
    # exit()
    for p in my_cmt.p:
        if diff_pos:
            if p + 1 not in inpos:
                keep_col.append(False)
            else:
                keep_col.append(True)
        else:
            if p not in inpos:
                keep_col.append(False)
            else:
                keep_col.append(True)
    keep_col = np.array(keep_col)
    my_cmt.filter_positions(keep_col)
    for p in inpos:
        if diff_pos:
            if p - 1 not in my_cmt.p:
                print('Pos not consistent! Exit!')
                exit()
        else:
            if p not in my_cmt.p:
                print('Pos not consistent! Exit!')
                exit()
    '''
    

    combined_array = reorder_norm(combined_array, my_cmt)
    #### Remove low quality samples
    #print(combined_array.shape,len(p))
    #print(p,len(p))
    #exit()
    if not min_cov_samp==100:
        combined_array,p=remove_low_quality_samples(combined_array, min_cov_samp,p)
    #print(np.where(p==864972))
    #exit()
    #### Remove bad positions
    #pfgap=[]
    combined_array,p,pfgap,praw=remove_lp(combined_array,p,my_cmt,my_calls,median_cov )
    dgap={}
    #praw=p
    for s in praw:
        if s not in pfgap:
            dgap[s]='0'
        else:
            dgap[s]='1'
    

    return combined_array,p,dgap


def load_test_name(infile):
    dt={}
    f=open(infile,'r')
    while True:
        line=f.readline().strip()
        if not line:break
        dt[line]=''
    return dt

def CNN_predict(data_file_cmt,data_file_cov,out,samples_to_exclude,min_cov_samp):
    if not os.path.exists(out):
        os.makedirs(out)
    setup_seed(1234)
    # dr=load_test_name('../39features-train.txt')
    # dt=load_test_name('../39features-test.txt') # test dict
    #indir='CNN_select_10features_mask_balance_no_Sep'
    #indir='CNN_select_10features_mask_balance_align_slides'
    ########## Kcp paper ###########
    #indir='CNN_select_40features_science_kcp' # these datasets have correct labels now!!!
    # indir='../../../Other_datasets/ScienceTM_KCp/CNN_select_features_target_10features_kcp'
    # in_npz='../../../Other_datasets/ScienceTM_KCp/npz_files_from_server'
    ########## elife-Sau paper ###########
    # indir='../../../Other_datasets/eLife-Sau-2022/CNN_select_features_target_10features_sau'
    # in_npz='../../../Other_datasets/eLife-Sau-2022/npz_files_from_server'
    ########## PNAS-Sau paper ###########
    # indir='../../../Other_datasets/PNAS-Sau-2014/CNN_select_features_target_10features_sau'
    # in_npz='../../../Other_datasets/PNAS-Sau-2014/npz_files_from_server'
    incount = data_file_cmt
    incov = data_file_cov
    '''
    for filename in os.listdir(in_npz_dir):
        # if re.search('DS',filename):continue
        # data = np.load(indir + '/' + filename + '/cnn_select.npz')
        # pre=re.sub('_Bfg','',filename)
        # pre=re.sub('_Cae','',pre)
        # pre = re.sub('_Sep', '', pre)
        # ind=in_npz+'/'+pre
        if re.search('mutation',filename):
            incount=in_npz_dir+'/'+filename
            pre=re.sub('_candidate_mutation_table.npz','',filename)
            pre=re.sub('group_','',pre)
        if re.search('coverage_matrix_raw',filename):
            incov=in_npz_dir+'/'+filename
    '''
    if not os.path.exists(incount) or not os.path.exists(incov):
        print('Mutation table or coverage matrix is not available! Please check! Exit.')
        exit()
    ########## From Collaborator data - Pseudomonas_aeruginosa_P-12 #######
    #mut='/Users/liaoherui/Downloads/clabsi_project/data/candidate_mutation_tables/group_29_candidate_mutation_table.pickle.gz'
    mut=incount
    #cov='/Users/liaoherui/Downloads/clabsi_project/data/candidate_mutation_tables/group_29_coverage_matrix_raw.pickle.gz'
    cov=incov
    #pre='Ecoli-P33'
    #pre='Kpn-P4'
    #pre='Pae-P14'
    fig_odir=out
    # indir='CNN_select_features'
    #train_datasets=[]
    test_datasets=[]
    #dsize=[]

    info=[]
    #print('Test data :'+pre)
    odata,pos,dgap=data_transform(mut,cov,fig_odir,samples_to_exclude,min_cov_samp)
    #odata=odata[np.where(pos==864972)]
    #odata=odata[:,20:23,:,:]
    #print(odata,odata.shape)
    #exit()
    #print(odata.shape,pos)
    #exit()
    #print(odata[5])
    #exit()
    print('Transformed data shape:',odata.shape)
    #test_datasets.append((data['x'][:,:,8:].astype(np.float64), data['label']))
    nlab=np.zeros(odata.shape[0])
    #test_datasets.append((odata))
    test_datasets.append((odata,nlab))
    #odata[2][5][2][1]=1
    #print(odata.shape)
    #exit()
    #exit()
    #test_datasets.append((data['x'][:, :, index].astype(np.float64), data['label']))
    for p in pos:
        info.append(str(p))
    #dsize.append(len(nlab))
    #return info

    info = np.array(info)
    #dsize=np.array(dsize)
    #exit()
    #train_datasets=np.delete(datasets, selected_indices, axis=0)
    #train_datasets=datasets[keep]
    #test_datasets = datasets[selected_indices]

    # #### stat datasets info
    # def stat(datasets,pre):
    #     p=0
    #     n=0
    #     for i,(data,label) in enumerate(datasets):
    #         n+=np.count_nonzero(label == 0)
    #         p+=np.count_nonzero(label == 1)
    #     print(pre,' dataset has ', len(datasets),' lineages, ',p,' true SNPs, ',n,' false SNPs, total', n+p,' SNPs',flush=True)
    #
    # #stat(train_datasets,'Training')
    # stat(test_datasets,'Test')
    # #exit()


    #Create DataLoaders
    def create_dataloader(datasets):
        dataloader=[]
        for i, (data,label)  in enumerate(datasets):
            dataset = CustomDataset(data, label)
            #dataset2 = CustomDataset(data2, labels2)
            dataloader_tem = DataLoader(dataset, batch_size=512, shuffle=False)
            #dataloader2 = DataLoader(dataset2, batch_size=32, shuffle=True)
            dataloader.append(dataloader_tem)
        return dataloader

    # #train_loader=create_dataloader(train_datasets)
    test_loader=create_dataloader(test_datasets)


    # Training
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    print('The device you are using is: ',device,flush=True)
    #model = CNNModel_2(n_channels=11).to(device)
    model= CNNModel(n_channels=4).to(device)

    model.load_state_dict(torch.load(script_dir+'/CNN_models/checkpoint_best_3conv.pt'))
    
    # #weight = torch.tensor([0.01, 0.99])
    # #criterion = nn.BCEWithLogitsLoss(weight=weight)
    # optimizer = optim.Adam(model.parameters(), lr=0.001)
    # early_stopping = EarlyStopping(patience=20, verbose=True)

    # valid_losses=[]
    # num_epochs = 500
    #o=open('check_res_cnn_39features_multichannel_science_kcp_with_reorder_split_m2_basechannel_large_remove.txt','w+')
    #o=open('check_res_cnn_39features_multichannel_elife_sau_with_reorder_split_m2_basechannel_large_remove.txt','w+')
    #o=open('check_res_cnn_39features_multichannel_pnas_sau_with_reorder_split_m2_basechannel_large_remove.txt','w+')
    #o=open('check_res_cnn_39features_multichannel_elife_sau_with_mask.txt','w+')
    #o=open(out+'/cnn_res.txt','w+')
    #o.write('Pos_info\tPredicted_label\tProbability\tGap_filt\n')

    #print('Train')
    model.eval()

    predictions = []
    y_test=[]
    with torch.no_grad():
        for loader in test_loader:
            for inputs,label in loader:
                #print(inputs)
                #exit()
                inputs=inputs.to(device)
                #inputs = torch.from_numpy(np.float32(inputs)).to(device)
                # print(inputs[20])
                # exit()
                outputs = model(inputs)
                # print(model(inputs))
                # exit()
                predictions.extend(outputs.cpu().numpy().flatten())
    #loss = criterion(outputs, y_test)
    # Convert predictions to binary labels
    prob=np.array(predictions)
    predictions = (np.array(predictions) > 0.5).astype(int)
    #print(predictions,prob)
    #exit()
    # print(predictions)
    y_pred = predictions
    #c=0
    #print('Predicted results:',np.count_nonzero(y_pred),' true SNPs, ',len(y_pred)-np.count_nonzero(y_pred),' false SNPs.')
    '''
    for s in y_pred:
        o.write(info[c]+'\t'+str(s)+'\t'+str(prob[c])+'\t'+dgap[int(info[c])]+'\n')
        c+=1
    '''
    # Return all reamining positions and CNN's predicted probabilities array
    return pos,y_pred,prob,dgap
    # accuracy = accuracy_score(y_test, y_pred)
    # # Calculate precision
    # precision = precision_score(y_test, y_pred)
    #
    # # Calculate recall (sensitivity)
    # recall = recall_score(y_test, y_pred)
    # # Calculate F1-score
    # f1 = f1_score(y_test, y_pred)
    # #roc_auc = roc_auc_score(y_test, y_pred)
    # print('Test dataset accuracy is:', accuracy, ' precision:', precision, ' recall:', recall, ' f1-score:', f1, flush=True)
    # print('Test dataset accuracy is:', accuracy, ' precision:', precision, ' recall:', recall, ' f1-score:', f1, flush=True,file=o)
    # conf_matrix = confusion_matrix(y_test, y_pred)
    # class_report = classification_report(y_test, y_pred)
    # print('Confusion matrix:', conf_matrix, '\nClassification report:', class_report)
    #
    #
    # #print(new_data.shape,len(new_data))
    # #torch.save(model.state_dict(),'cae_7000plus_cnn_af.pt')

# Herui's test
#pos,pred,prob=CNN_predict('../npz_of_Test_data/cae_pe/candidate_mutation_table','cae_pe_local')
#print(pos,pred,prob)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  7 20:00:41 2022

@author: evanqu
"""
import numpy as np
import pickle
import argparse
import gzip
import gus_helper_functions as ghf

#%%
def chrpos2index(chrpos,chr_starts):
    '''Python version of chrpos2index.m

    Args:
        chrpos (arr): px2 array of position and chromsome idx.
        chr_starts (arr): Vector of chromosome starts (begins at 0).

    Returns:
        p (arr): Vector of position indexes.

    '''
    #if np.size(chrpos,0) < np.size(chrpos,1):
    #   chrpos=chrpos.T
    #    print('Reversed orientation of chrpos')
        
    if len(chr_starts) == 1:
        p=chrpos[:,1]
    else:
        p=chr_starts[chrpos[:,0]-1]+chrpos[:,1]

    return p

def generate_positions_snakemake(positions_files_list, REFGENOMEDIRECTORY):
    '''Python version of generate_positions_snakemake.m
    
    Args:
        paths_to_input_p_files (list): List of input positions files.
        REFGENOMEDIRECTORY (str): Path to reference genome.

    Returns:
        combined_pos (arr): Vector of variable positions across samples.

    '''
    
    [chr_starts,genome_length,scaf_names] = ghf.genomestats(REFGENOMEDIRECTORY)
    
    # initialize vector to count occurrances of variants across samples
    timesvariant = np.zeros((genome_length,1))
    
    for i in range(len(positions_files_list)):
        #load in positions array for sample
        with gzip.open(positions_files_list[i].rstrip('\n'),"rb") as f:
            positions=pickle.load(f)
        if positions.shape[0]==0:continue
        #if len(positions)>2:

        x=chrpos2index(positions,chr_starts)
        #if 307924 in positions:
        #    print(positions_files_list[i],x)            
        timesvariant[x]=timesvariant[x]+1
    
    #print(timesvariant[307924])
    #Keep positions that vary from the reference in at least one sample but
    #that don't vary from the reference in ALL samples
    #combined_pos = np.where((timesvariant > 0) & (timesvariant < len(positions_files_list)))[0]
    combined_pos = np.where((timesvariant > 0))[0]
    #print(307924 in combined_pos)
    
    return combined_pos
    

def combine_positions(path_to_positions_files, path_to_output_p_file, path_to_outgroup_boolean_file, REFGENOMEDIRECTORY):
    
    #in_outgroup: booleans giving whether or not each sample is in the outgroup
    print("Processing outgroup booleans...")
    in_outgroup=[]
    with open(path_to_outgroup_boolean_file) as file:
        for line in file:
            in_outgroup.append(line)
    #Bool of samples to include
    include = [not i for i in in_outgroup]
    
    #Get positions on reference genome
    [chr_starts,genome_length,scaf_names] = ghf.genomestats(REFGENOMEDIRECTORY)
    
    #Find positions with at least 1 fixed mutation relative to reference genome
    print('\n\nFinding positions with at least 1 fixed mutation...\n')
    
    positions_files_ls=[]
    with open(path_to_positions_files) as file:
        for line in file:
            positions_files_ls.append(line)
            
    # print(f"\nIngroup paths used to generate positions: {positions_files_ls[include]}")
    print(include)
    
    cp = generate_positions_snakemake(positions_files_ls,REFGENOMEDIRECTORY)
    print(f"Found {len(cp)} positions where provided vcfs called a fixed variant in at least one in-group sample \n")

    #Todo: Add candidate positions manually
    #Combine different types of positions
    allp=cp
    
    print("Saving list of all positions...")
    with open(path_to_output_p_file,"wb") as wf:
        pickle.dump(allp,wf)
    
    return

#%%
if __name__ == '__main__':
    
    # SCRIPTS_DIR="scripts"
    # sys.path.insert(0, SCRIPTS_DIR)
        
    parser = argparse.ArgumentParser()
    
    parser.add_argument('-i', type=str, help='Path to List of input positions files',required=True)
    parser.add_argument('-r', type=str, help='Path to reference genome directory',required=True)
    parser.add_argument('-o', type=str, help='Path to output positions file', required=True)
    parser.add_argument('-b', type=str, help='Outgroup boolean', required=True)
    
    args = parser.parse_args()
    
    print()

    combine_positions(args.i,args.o,args.b,args.r)



import runpy


def main():
    """Entry point for running the downstream analysis script."""
    runpy.run_module("accusnv.accusnv_downstream", run_name="__main__")


if __name__ == "__main__":
    main()
